///
/// \note	
/// 		This sourcefile is provided for free use for non-commercial purposes only. The
/// 		Software is provided 'as is' and no responsibility can be taken for any damages
/// 		that may be caused while using it. If you make use of this software within 
/// 		your code then it would be nice to have a small credit somewhere in your
/// 		finished product for my hard work :(
/// 		An e-mail saying you found it useful and possibly even a link to your product
/// 		would be really nice too ;)
///
///			Slip me a couple of quid and i might be kind enough to offer assistance in
/// 		sorting out your exporters ;) 
///
///				(C) 2004 Rob 'give me a job offer' Bateman 

#ifdef WIN32
	#ifndef NT_APP
		#define NT_APP
	#endif
	
	#pragma warning(disable:4786)
#endif

#include <algorithm>
#include <iostream>
#include <fstream>
#include <vector>
#include <string>
#include <map>

#include <maya/MAnimControl.h>
#include <maya/MVector.h>
#include <maya/MQuaternion.h>
#include <maya/MDagPath.h>
#include <maya/MEulerRotation.h>
#include <maya/MFileIO.h>
#include <maya/MMatrix.h>
#include <maya/MFnTransform.h>
#include <maya/MFnIkJoint.h>
#include "Bone.h"	



/// this is a global variable that is used to generate unique bone
/// id numbers. 
/// 
unsigned int BoneRefID = 0;

/// \brief	ctor
/// \param	name	-	the maya name of the bone
/// 
CBoneRef::CBoneRef(std::string n) 
	: name(n),m_Parent(0),id(++BoneRefID) 
{
}

/// \brief	ctor
/// \param	name	-	the maya name of the bone
/// \param	path	-	the path through maya's DAG to the node
/// 
CBoneRef::CBoneRef(std::string n,MDagPath &p)
	: name(n),path(p),m_Parent(0),id(++BoneRefID)
{
}

/// this is a global variable that is used to generate unique object
/// id numbers. 
/// 
unsigned int ObjectRefID = 0;

/// all of the object references are stored within a map
typedef std::map<std::string, unsigned int> ObjectMap;

/// the iterator type
typedef ObjectMap::iterator ObjectIT;

/// a map of all useful transforms
ObjectMap g_Objects;


/// \brief	Each object is associated with a unique objectID. This function will return
///			a unique integer key for each string passed to it. 
///			If the key already exists it returns the old one, otherwise it creates
///			a new object entry.
/// 
unsigned int GetObjectID(const std::string name)
{
	// find the bone
	ObjectIT it = g_Objects.find(name);

	// if not found
	if (it == g_Objects.end()) 
	{
		// insert entry into bone map
		std::pair<ObjectIT,bool> pair_ = g_Objects.insert( std::make_pair( name , ++ObjectRefID  )  );
		
		return pair_.first->second;
	}
	return it->second;
}



/// the bones are stored in a map for quick loop up
typedef std::map<std::string, CBoneRef> BoneMap;

/// the bone iterator type
typedef BoneMap::iterator BoneIT;

/// an array of important transform nodes
std::vector<unsigned int> g_ImportantTransforms;

/// a map of all useful transforms
BoneMap g_Bones;


/// \brief	This function parents the specified bone (NOT SHAPE!) to the
///			parent bone.
/// \param	child	-	the child bone ID
/// \param	bone	-	the parent bone ID
/// 
void AddChildToBone(unsigned int _child,unsigned int bone_id)
{
	// dont want to parent to myself
	if (_child == bone_id) {
		return;
	}

	BoneIT it = g_Bones.begin();

	for(; it != g_Bones.end(); ++it) {
		if (bone_id == it->second.id) {
			break;
		}
	}

	if (it != g_Bones.end()) {
		CBoneRef& joint = it->second;

		std::sort( joint.m_Children.begin(), joint.m_Children.end() );

		std::vector<unsigned int>::iterator itChildren = joint.m_Children.begin();

		for( ; itChildren != joint.m_Children.end() ; ++ itChildren ) {
			if ( *itChildren == _child) 
			{
				return;				
			}
		}
		joint.m_Children.push_back(_child);
	}
}


/// \brief	This function parents the specified node (NOT TRANSFORM!) to the
///			parent bone.
/// \param	object	-	the object ID
/// \param	bone	-	the parent bone ID
/// 
void AddObjectToBone(unsigned int _child,unsigned int bone_id)
{
	BoneIT it = g_Bones.begin();

	for(; it != g_Bones.end(); ++it) {
		if (bone_id == it->second.id) {
			break;
		}
	}

	if (it != g_Bones.end()) {
		CBoneRef& joint = it->second;

		std::sort( joint.m_ChildObjects.begin(), joint.m_ChildObjects.end() );

		std::vector<unsigned int>::iterator itChildren = joint.m_ChildObjects.begin();

		for( ; itChildren != joint.m_ChildObjects.end() ; ++ itChildren ) {
			if ( *itChildren == _child) 
			{
				return;				
			}
		}
		joint.m_ChildObjects.push_back(_child);
	}
}



/// \brief	When we decide that we are going to output a specific mesh, light or other object,
///			we call this function which recursively ascends the hierarchy until it reachs the root
///			and then recursively decends into the hiereachy building up references of all nodes 
///			that it finds.
/// \param	obj	-	the joint to build a reference for
/// \param	recursive	-	should the function recursivel check all children?
/// \return	the joint ID of the joint queried, or zero if failed
/// 
unsigned int BuildJointRef(const MObject &obj,bool doChildren)
{
	if (obj.hasFn(MFn::kTransform)) {
		// set function set to node
		MFnTransform fnX(obj);

		// find the bone
		BoneIT it = g_Bones.find(fnX.name().asChar());

		// if not found
		if (it == g_Bones.end()) 
		{
			// get path to bone
			MDagPath p;
			fnX.getPath(p);

			// insert entry into bone map
			std::pair<BoneIT,bool> pair_ = g_Bones.insert( std::make_pair( std::string( fnX.name().asChar() ), CBoneRef(fnX.name().asChar(),p))  );
			
			// get ref to joint
			CBoneRef& joint = pair_.first->second;

			joint.name = fnX.name().asChar();

			//////////////////////////////////////////////////////////////////////////
			
			{
				if (fnX.parentCount() != 0) 
				{
					joint.m_Parent = BuildJointRef(fnX.parent(0),doChildren);
				}
				else
				{
					joint.m_Parent = 0;
					g_ImportantTransforms.push_back(joint.id);
				}
			}

			//////////////////////////////////////////////////////////////////////////
			
			if(doChildren)
			{
				for(int i=0;i!=fnX.childCount();++i)
				{
					MObject oChild = fnX.child(i);
					if (oChild.hasFn(MFn::kTransform)) {
						unsigned int jref = BuildJointRef(oChild);
						if(jref!=0)
							joint.m_Children.push_back( jref );
					}
					else
					if ( oChild.hasFn(MFn::kMesh) || 
						 oChild.hasFn(MFn::kLight) ) 
					{
						MFnDependencyNode fnDep(oChild);
						unsigned int jref = GetObjectID(fnDep.name().asChar());
						AddObjectToBone(jref,joint.id);
					}
				}
			}
			
			//////////////////////////////////////////////////////////////////////////

			return joint.id;
		}
		return it->second.id;
	}
	return 0;
}

/// \brief	This function outputs a specific bone into the file. 
/// \param	ofs		-	the outputfile stream
/// \param	bone	-	the bone to output
void OutputTransform(std::ofstream& ofs,CBoneRef& bone)
{
	double Scale[3];
	MEulerRotation EulerRotation;
	MQuaternion JO(0,0,0,1);
	MQuaternion RO(0,0,0,1);
	MQuaternion R(0,0,0,1);

	// set function set to node
	MFnTransform fnX(bone.path);

	MMatrix mat = fnX.transformation().asMatrix();
	MMatrix inverse = mat.inverse();

	fnX.getScale(Scale);
	fnX.getRotation (R);

	if( fnX.object().hasFn(MFn::kJoint) )
	{
		MFnIkJoint fnJ(fnX.object());
		fnJ.getOrientation(JO);
	}

	RO = fnX.rotateOrientation(MSpace::kObject);

	MVector trans = fnX.translation(MSpace::kTransform );

	ofs << "BONE " << bone.name << std::endl;
	ofs << "\tid " << bone.id << std::endl;
	ofs << "\tparent " << bone.m_Parent << std::endl;
	ofs << "\tChildren " << static_cast<unsigned int>(bone.m_Children.size()) << "\t";
	std::vector<unsigned int>::iterator it = bone.m_Children.begin();
	for( ; it != bone.m_Children.end(); ++it )
	{
		ofs << *it << " ";
	}
	ofs << std::endl;	
	ofs << "\tParentedNodes " << static_cast<unsigned int>(bone.m_ChildObjects.size()) << "\t";
	it = bone.m_ChildObjects.begin();
	for( ; it != bone.m_ChildObjects.end(); ++it )
	{
		ofs << *it << " ";
	}
	ofs << std::endl;

	ofs << "\tScale\t" 			<< (float)Scale[0] 	<< " " << (float)Scale[1] 	<< " " << (float)Scale[2] 	<< std::endl;
	ofs << "\tTranslation\t" 	<< (float)trans.x 	<< " " << (float)trans.y 	<< " " << (float)trans.z 	<< std::endl;
	ofs << "\tRotationOrient\t" << (float)RO.x 		<< " " << (float)RO.y 		<< " " << (float)RO.z 		<< " " << (float)RO.w << std::endl;
	ofs << "\tRotation\t" 		<< (float)R.x 		<< " " << (float)R.y 		<< " " << (float)R.z 		<< " " << (float)R.w  << std::endl;
	ofs << "\tJointOrient\t" 	<< (float)JO.x 		<< " " << (float)JO.y 		<< " " << (float)JO.z 		<< " " << (float)JO.w << std::endl;

	// output a hopefully world space matrix :)
	ofs << "\tBaseMatrix\t" << (float)mat(0,0) << " " << (float)mat(0,1) << " " << (float)mat(0,2) << " " << (float)mat(0,3) << " "
	                        << (float)mat(1,0) << " " << (float)mat(1,1) << " " << (float)mat(1,2) << " " << (float)mat(1,3) << " "
							<< (float)mat(2,0) << " " << (float)mat(2,1) << " " << (float)mat(2,2) << " " << (float)mat(2,3) << " "
							<< (float)mat(3,0) << " " << (float)mat(3,1) << " " << (float)mat(3,2) << " " << (float)mat(3,3) << "\n";
}


/// \brief	This function outputs all of the transforms in the scene into the file
/// \param	ofs	-	the output file stream
/// 
void OutputTransforms(std::ofstream& ofs)
{
	/// a map of all useful transforms
	BoneIT it = g_Bones.begin();
	for( ; it != g_Bones.end(); ++it )
	{
		std::cout << "outputting xform: " << it->first << std::endl;
		OutputTransform(ofs,it->second);
	}

}


/// \brief	This function outputs all animation data in the cycle between the bounds
///			specified in the time slider when the file was saved.
/// \param	ofs	-	the output file stream
/// 
void OutputAnimation(std::ostream &ofs)
{
	ofs << "ANIMCYCLE " 
		<< MFileIO::currentFile().asChar() 
		<< "\n";
	
	// get start and end times
	int StartTime = static_cast<int>(MAnimControl::minTime().as(MTime::kPALFrame));
	int EndTime   = static_cast<int>(MAnimControl::maxTime().as(MTime::kPALFrame));
	float fStartTime = static_cast<float>(MAnimControl::minTime().as(MTime::kSeconds));
	float fEndTime   = static_cast<float>(MAnimControl::maxTime().as(MTime::kSeconds));

	ofs << "\tnumbones " << static_cast<unsigned int>(g_Bones.size()) << std::endl;
	ofs << "\tstart " << (float)fStartTime << std::endl;
	ofs << "\tend "   << (float)fEndTime   << std::endl;
	ofs << "\tnum_frames " << (EndTime-StartTime+1) << std::endl;

	/// a map of all useful transforms
	BoneIT it = g_Bones.begin();
	for( ; it != g_Bones.end(); ++it )
	{
		// attach function set
		MFnTransform fnX(it->second.path);

		// output bone chunk
		ofs << "\tBONEANIM " << it->second.id << std::endl;
		std::cout << "outputting bone anim " << it->first << std::endl;

		for(int i=StartTime;i<=EndTime;++i)
		{
			// set the current time
			MAnimControl::setCurrentTime(MTime(i,MTime::kPALFrame));

			// get rotation
			MQuaternion quat;
			fnX.getRotation(quat);

			// get translation
			MVector trans = fnX.translation(MSpace::kTransform );

			// get scale
			double scale[3];
			fnX.getScale(scale);

			// output keyframe
			ofs << "\t\t" << (float)quat.x   << " "  << (float)quat.y   << " " << (float)quat.z   << " " << (float)quat.w 
				<< "\t" << (float)trans.x  << " "  << (float)trans.y  << " " << (float)trans.z 
				<< "\t" << (float)scale[0] << " "  << (float)scale[1] << " " << (float)scale[2] << std::endl;
		}
	}
}






