//////////////////////////////////////////////////////////////////////////
/// 
/// \file	Bone.h
/// \author	Rob Bateman
///	\brief	This file defines the mechanism for exporting bones and 
///			transform information from the scenegraph of maya. In Maya all
///			shapes (eg meshes, curves, surfaces, cameras, lights etc) are
///			placed in the scene via the use of a trasnform node. Infact,
///			most of the animation data etc will be retrieved from the 
///			transform nodes so they are fairly useful. The only real 
///			with Maya is that it tend to have loads of fairly useless 
///			transforms floating around for things such as the default cameras
///			3D and projection texture placements and a whole host of other 
///			rubbish. To circumvent this, every time we read an important
///			node such as a mesh, I recursively read back from the node until
///			all transforms in the hierarchy are found. This means that only 
///			the transform hierarchies of the correct bones are found. 
///			Theoretically that should work..... 
///			The bones themseleves are referenced using ID numbers which is
///			probably a bad idea but never mind....
///
/// \note	
/// 		This sourcefile is provided for free use for non-commercial purposes only. The
/// 		Software is provided 'as is' and no responsibility can be taken for any damages
/// 		that may be caused while using it. If you make use of this software within 
/// 		your code then it would be nice to have a small credit somewhere in your
/// 		finished product for my hard work :(
/// 		An e-mail saying you found it useful and possibly even a link to your product
/// 		would be really nice too ;)
///
///			Slip me a couple of quid and i might be kind enough to offer assistance in
/// 		sorting out your exporters ;) 
///
///				(C) 2004 Rob 'give me a job offer' Bateman 
/// 
//////////////////////////////////////////////////////////////////////////

#ifdef WIN32
#pragma once
#pragma warning(disable:4786)
#endif

#ifndef BONE_H_INCLUDED_
#define BONE_H_INCLUDED_

#include <maya/MObject.h>
#include <maya/MDagPath.h>

#include <string>
#include <vector>
#include <fstream>

struct CBoneRef 
{
	CBoneRef(std::string n);
	CBoneRef(std::string n,MDagPath &p);

	/// the name of the bone
	std::string name;

	/// the dag path to the node
	MDagPath    path;

	/// an array containing the names of lights, locator nodes, meshes etc
	std::vector<std::string> m_ParentedNodes;

	/// an array of child bones id's
	std::vector<unsigned int> m_Children;

	/// an array of child geometry id's
	std::vector<unsigned int> m_ChildObjects;

	/// the id of the parent node
	unsigned int m_Parent;

	/// the unique id of the bone
	unsigned int id;
};




/// \brief	Each object is associated with a unique objectID. This function will return
///			a unique integer key for each string passed to it. 
///			If the key already exists it returns the old one, otherwise it creates
///			a new object entry.
/// 
extern unsigned int GetObjectID(std::string);

/// \brief	This function outputs all animation data in the cycle between the bounds
///			specified in the time slider when the file was saved.
/// \param	ofs	-	the output file stream
/// 
extern void OutputAnimation(std::ostream&);

/// \brief	This function outputs all of the transforms in the scene into the file
/// \param	ofs	-	the output file stream
/// 
extern void OutputTransforms(std::ofstream&);

/// \brief	When we decide that we are going to output a specific mesh, light or other object,
///			we call this function which recursively ascends the hierarchy until it reachs the root
///			and then recursively decends into the hiereachy building up references of all nodes 
///			that it finds.
/// \param	obj	-	the joint to build a reference for
/// \param	recursive	-	should the function recursivel check all children?
/// \return	the joint ID of the joint queried
/// 
extern unsigned int BuildJointRef(const MObject&,bool = true);

/// \brief	This function parents the specified node (NOT TRANSFORM!) to the
///			parent bone.
/// \param	object	-	the object ID
/// \param	bone	-	the parent bone ID
/// 
extern void AddObjectToBone(unsigned int,unsigned int);

/// \brief	This function parents the specified bone (NOT SHAPE!) to the
///			parent bone.
/// \param	child	-	the child bone ID
/// \param	bone	-	the parent bone ID
/// 
extern void AddChildToBone(unsigned int,unsigned int);


#endif







