/////////////////////////////////////////////////////////////////////////////////////////
/// 
///	\file	Light.cpp
/// \author	Rob Bateman			( robthebloke@hotmail.com )
/// \brief	This file defines the single function OutputLight which will write the light
///			data to the file. 
///
/// \note	
/// 		This sourcefile is provided for free use for non-commercial purposes only. The
/// 		Software is provided 'as is' and no responsibility can be taken for any damages
/// 		that may be caused while using it. If you make use of this software within 
/// 		your code then it would be nice to have a small credit somewhere in your
/// 		finished product for my hard work :(
/// 		An e-mail saying you found it useful and possibly even a link to your product
/// 		would be really nice too ;)
///
///			Slip me a couple of quid and i might be kind enough to offer assistance in
/// 		sorting out your exporters ;) 
///
///				(C) 2004 Rob 'give me a job offer' Bateman 
///
/////////////////////////////////////////////////////////////////////////////////////////



#ifdef WIN32
	#ifndef NT_APP
		#define NT_APP
	#endif
	
	#pragma warning(disable:4786)
#endif
/////////////////////////////////////////////////////////////////////////////////////////

#include "Light.h"
#include "Bone.h"

#include <maya/MColor.h>
#include <maya/MMatrix.h>
#include <maya/MVector.h>

#include <maya/MFnAmbientLight.h>
#include <maya/MFnAreaLight.h>
#include <maya/MFnDagNode.h>
#include <maya/MFnDirectionalLight.h>
#include <maya/MFnLight.h>
#include <maya/MFnPointLight.h>
#include <maya/MFnSpotLight.h>
#include <maya/MFnTransform.h>

#include <iostream>
#include <cassert>

/////////////////////////////////////////////////////////////////////////////////////////
/// \brief	This function outputs light data to the specifed filestream. The MObject
///			passed as an argument is the handle to the maya node.
/// \param	ofs	-	the output stream
///	\param	obj	-	the maya object
/// 
void OutputLight(std::ofstream& ofs,MObject &obj)
{
	// make sure we only get valid nodes here
	assert( !obj.isNull() );

	// make sure the MObject represents a light!
	assert( obj.hasFn(MFn::kLight) );

	// attach a light function set to the object
	MFnLight fnLight(obj);

	// if the object is an intermediate object, we will ignore it
	if ( !fnLight.isIntermediateObject() ) 
	{
		//////////////////////////////////////////////////////////////////////////////////
		// sort out parenting and xform info
		//
		unsigned int iParent=0;
		if (fnLight.parentCount()>0) {
			iParent = BuildJointRef( fnLight.parent(0) );
		}
		unsigned int iNode =  GetObjectID(fnLight.name().asChar());

		// add the light to the joint ref so we know we are exporting it.
		AddObjectToBone(iNode,iParent);

		// get light color
		MColor color(0.0f,0.0f,0.0f);
		color = fnLight.color();

		// output some light info
		ofs << "LIGHT "

			// output name
			<< fnLight.name().asChar()
			<< "\n\tId "
			<< iNode
			<< "\n\tParent\t"
			<< iParent
			
			// output light type
			<< "\n\tType=\t\t" 
			<< obj.apiTypeStr()

			// output color 
			<< "\n\tColor=\t\t"
			<< (float)color.r
			<< " "
			<< (float)color.g
			<< " "
			<< (float)color.b 

			// output intensity
			<< "\n\tIntensity=\t" 
			<< (float)fnLight.intensity()

			<< std::endl;


		// if you need to get specific attributes associated with specific
		// light types then use something like this....
		//
		switch(obj.apiType()) 
		{
		case MFn::kPointLight:
			{
				MFnPointLight fnPoint(obj);

			}
			break;
		case MFn::kAmbientLight:
			{
				MFnAmbientLight fnAmbient(obj);

			}
			break;
		case MFn::kSpotLight:
			{
				MFnSpotLight fnSpot(obj);

			}
			break;
		case MFn::kDirectionalLight:
			{
				MFnDirectionalLight fnDir(obj);

			}
			break;
		case MFn::kAreaLight:
			{
				MFnAreaLight fnArea(obj);

			}
			break;
		default:
			break;
		}
	}
}

