/////////////////////////////////////////////////////////////////////////////////////////
///	\file	Mesh.cpp
/// \author	Rob Bateman
/// \brief	This file defines the single function OutputMesh which will write the mesh
///			data to the file. 
/// \note	When accessing any node data within maya, you must use a function set
///			and attach it to the relevant MObject within maya. All function sets
///			are listed with the prefix MFn. The class MFn also contains the type
///			enumerations for the various maya nodes, in the case of a mesh, 
/// 
///									MFn::kMesh
/// 
///
/// \note	
/// 		This sourcefile is provided for free use for non-commercial purposes only. The
/// 		Software is provided 'as is' and no responsibility can be taken for any damages
/// 		that may be caused while using it. If you make use of this software within 
/// 		your code then it would be nice to have a small credit somewhere in your
/// 		finished product for my hard work :(
/// 		An e-mail saying you found it useful and possibly even a link to your product
/// 		would be really nice too ;)
///
///			Slip me a couple of quid and i might be kind enough to offer assistance in
/// 		sorting out your exporters ;) 
///
///				(C) 2004 Rob 'give me a job offer' Bateman 
///			
/////////////////////////////////////////////////////////////////////////////////////////


#ifdef WIN32
	#ifndef NT_APP
		#define NT_APP
	#endif
	
	#pragma warning(disable:4786)
#endif

/////////////////////////////////////////////////////////////////////////////////////////
// includes
//
#include <maya/MFloatMatrix.h>
#include <maya/MFloatVector.h>
#include <maya/MFloatVectorArray.h>
#include <maya/MMatrix.h>
#include <maya/MPlug.h>
#include <maya/MPlugArray.h>
#include <maya/MPoint.h>
#include <maya/MPointArray.h>

#include <maya/MFnMesh.h>
#include <maya/MFnTransform.h>

#include <maya/MItDependencyNodes.h>
#include <maya/MItMeshPolygon.h>
#include <maya/MItMeshVertex.h>

#include <iostream>
#include <fstream>
#include <vector>
#include <string>
#include <map>

#include "Mesh.h"
#include "Bone.h"
#include "Material.h"


/////////////////////////////////////////////////////////////////////////////////////////

/// \brief	This function outputs the specified mesh to the specified file stream
/// 
void OutputMesh(std::ostream& ofs,const MObject &obj)
{
	// attach a function set to the mesh node. We access all of the meshes data 
	// through the function set
	// 
	MFnMesh fnMesh(obj);



	// Maya has a tendancy of storing additional nodes and objects all over the place.
	// We may come accross meshes that are nodes used purely as history or other 
	// construction data. In our case, we want to ignore all 'intermediate objects'
	//
	if ( !fnMesh.isIntermediateObject() ) 
	{
		unsigned int iParent = BuildJointRef( fnMesh.parent(0) );
		unsigned int iNode =  GetObjectID(fnMesh.name().asChar());

		AddObjectToBone(iNode,iParent);

		// write out some info about the mesh
		// 
		ofs << "MESH "
			<< fnMesh.name().asChar() 
			<< "\n\tID= "
			<< GetObjectID(fnMesh.name().asChar())
			<< "\n\tParentID= "
			<< iParent
			<< "\n\tVertexCount= "
			<< fnMesh.numVertices() 
			<< "\n\tNormalCount= "
			<< fnMesh.numNormals()
			<< "\n\tTexCoordCount= "
			<< fnMesh.numUVs()
			<< "\n\tFaceCount= "
			<< fnMesh.numPolygons()
			<< std::endl;

		int i;


		// this will hold the returned vertex positions
		MPointArray vertices;
	
		// use the function set to get the points 
		fnMesh.getPoints(vertices);


		// go through each vertex in the array and write it to the file 
		//
		ofs << "\n\tVertices\n";
		for(i=0;i!=vertices.length() ; ++i)
		{					
			// output 
			//
			ofs << "\t\t" << (float)vertices[i].x << " " << (float)vertices[i].y << " " << (float)vertices[i].z << "\n";
		}



		// this will store the normals
		MFloatVectorArray normals;

		// get the normals
		fnMesh.getNormals(normals);

		// go through each normal in the array and write it to the file 
		//
		ofs << "\n\tNormals\n";
		for(i=0;i!=normals.length() ; ++i)
		{

			// transform the normal into world space
			//
			MFloatVector OutNormal = normals[i];

			// we must normalise the vector to counter any scaling in the transformation
			//
			OutNormal.normalize();

			// output the normal 
			// 
			ofs << "\t\t" << (float)OutNormal.x << " " << (float)OutNormal.y << " " << (float)OutNormal.z << "\n";
		}

		// these two arrays will hold the u & v texture co-ordinates
		//
		MFloatArray u_coords;
		MFloatArray v_coords;

		// get the uv coordinates
		//
		fnMesh.getUVs(u_coords,v_coords);

		if(u_coords.length())
		{
			ofs << "\n\tTexCoords\n";
			// go through each uv in the array and write it to the file 
			//
			for(i=0;i!=u_coords.length() ; ++i)
			{
				ofs << "\t\t" << (float)u_coords[i] << " " << (float)v_coords[i] << "\n";
			}
		}

		// this will hold references to the shaders used on the meshes
		MObjectArray Shaders;

		// this is used to hold indices to the materials returned in the object array
		MIntArray    FaceIndices;

		int k = 0;

		// this function builds a list of all surface shaders used on the mesh. Note
		// that the nodes returned are the shading groups, not the material nodes!
		fnMesh.getConnectedShaders(0,Shaders,FaceIndices);

		// having recieved the shader information, we will try to process the returned
		// shaders to find their name and ultimately find the relevant material ID that
		// references the list we have built up.
		std::vector< unsigned int > ShaderIDs;

		// start the first shader as a default zero value
		ShaderIDs.push_back(0);

		// for each surface shader
		for(k=0;k!=Shaders.length();++k)
		{
			// attachy function set...
			MFnDependencyNode fnShader(Shaders[k]);

			// get access to the surfaceShader attribute. This will be connected to
			// lambert , phong nodes etc.
			MPlug sshader = fnShader.findPlug("surfaceShader");

			MPlugArray materials;

			// get the material connected to the surface shader
			sshader.connectedTo(materials,true,true);

			// 
			for(unsigned int l=0;l<materials.length();++l)
			{
				MFnDependencyNode fnMat(materials[l].node());
				ShaderIDs.push_back( GetMaterialID(fnMat.name()) );
			}
		}



		k = 0;


		// to get at the face data, we need to use a mesh polygon iterator.
		//
		MItMeshPolygon itPoly(fnMesh.object());

		// keep looping until the iterator has finished
		// 
		while ( !itPoly.isDone() )
		{

			// get the number of vertices in this face
			unsigned int vcount = itPoly.polygonVertexCount();

			int matInd;
				{
					matInd =  ShaderIDs[ FaceIndices[k]+1 ];
				}

			// output some info about the face
			//
			ofs << "\n\tFACE "
				<< vcount 
				<< "\n\t\tMat= "
				<< matInd
				<< std::endl;

			// for each vertex on the face....
			//
			for(i = 0; i!= vcount; ++i)
			{
				// have to get the uv index seperately
				int uv_index;
				itPoly.getUVIndex(i,uv_index);

				// output indices for this vertex
				ofs << "\t\t" << itPoly.vertexIndex(i) << " " << itPoly.normalIndex(i);

				if(u_coords.length())
				{
					ofs << " " << uv_index;
				}
				ofs << "\n";
			}

			// move to next poly in the mesh
			//
			itPoly.next();
		}
	}
}
