//////////////////////////////////////////////////////////////////////////////////////////////
/// \file	NurbsSurface.cpp
/// \author	Robert Bateman
/// \brief	Might as well add support for nurbs curves, not the quickest of rendererable
/// 		primitve, though can be used for motion paths, camera splines etc.
/// \note	
/// 		This sourcefile is provided for free use for non-commercial purposes only. The
/// 		Software is provided 'as is' and no responsibility can be taken for any damages
/// 		that may be caused while using it. If you make use of this software within 
/// 		your code then it would be nice to have a small credit somewhere in your
/// 		finished product for my hard work :(
/// 		An e-mail saying you found it useful and possibly even a link to your product
/// 		would be really nice too ;)
///
///			Slip me a couple of quid and i might be kind enough to offer assistance in
/// 		sorting out your exporters ;) 
///
///				(C) 2004 Rob 'give me a job offer' Bateman 
///
//////////////////////////////////////////////////////////////////////////////////////////////

#ifdef WIN32
	#define NT_APP
#endif

#include "Bone.h"
#include "NurbsSurface.h"
#include "Material.h"
#include <maya/MPoint.h>
#include <maya/MPointArray.h>
#include <maya/MFnNurbsCurve.h>
#include <maya/MObjectArray.h>
#include <maya/MPlugArray.h>
#include <maya/MPlug.h>


unsigned int FindNurbsMaterial(const MObject& nurb)
{	

	MFnDependencyNode fn(nurb);
	MPlugArray plugs;

	// we are looking for a connection to a surface shader of some sort
	fn.getConnections(plugs);

	for(unsigned int i=0;i<plugs.length();++i)
	{
		MPlugArray OtherSide;
		plugs[i].connectedTo(OtherSide,false,true);
		for(unsigned int j=0;j<OtherSide.length();++j)
		{
			if ( OtherSide[j].node().hasFn(MFn::kShadingEngine) ) {
				MFnDependencyNode fnShader(OtherSide[j].node());

				MPlug matPlug = fnShader.findPlug("surfaceShader");

				MPlugArray MaterialConnections;
				matPlug.connectedTo(MaterialConnections,true,false);

				if (MaterialConnections.length()) {
					MFnDependencyNode fnMat(MaterialConnections[0].node());
					return GetMaterialID(fnMat.name());
				}
			}
		}
	}
	return 0;
}

void OutputNurbsSurface(std::ostream &ofs,const MObject& obj)
{
	MFnNurbsSurface fnSurface(obj);

	if (fnSurface.isIntermediateObject())
	{
		return;
	}

	int i;
	unsigned int iParent = BuildJointRef( fnSurface.parent(0) );
	unsigned int iNode =  GetObjectID(fnSurface.name().asChar());
	unsigned int iDegU = fnSurface.degreeU();
	unsigned int iDegV = fnSurface.degreeV();
	unsigned int iCVU  = fnSurface.numCVsInU();
	unsigned int iCVV  = fnSurface.numCVsInV();
	unsigned int iMat  = FindNurbsMaterial( obj );

	unsigned short formu=0,formv=0;

	switch(fnSurface.formInU()) {
	case MFnNurbsSurface::kClosed :
		formu=1;
		break;
	case MFnNurbsSurface::kPeriodic :
		formu=2;
		break;
	default:
		break;
	}

	switch(fnSurface.formInV()) {
	case MFnNurbsSurface::kClosed :
		formv=1;
		break;
	case MFnNurbsSurface::kPeriodic :
		formv=2;
		break;
	default:
		break;
	}

	AddObjectToBone(iNode,iParent);


	//	if degree is 3 in both directions, both curves are open and the cv counts 
	//	are 4 in both directions, then we have bezier patch
	//
	if( iDegU == 3 && 
		iDegV == 3 && 
		iCVU  == 4 && 
		iCVV  == 4 &&
		formu == 0 &&
		formv == 0 )
	{
		ofs << "BEZIER_PATCH " << fnSurface.name().asChar() 
			<< "\n\tID "		<< iNode
			<< "\n\tParentID "	<< iParent
			<< "\n\tmaterial "  << iMat;

		MPointArray pnts;
		fnSurface.getCVs(pnts);

		for(i=0;i<pnts.length();++i)
		{
			ofs << "\n\t\t" << (float)pnts[i].x 
				<< " "		<< (float)pnts[i].y 
				<< " "		<< (float)pnts[i].z;
		}

		ofs << "\n";
	}
	else
	{
		
		ofs << "NURBS_SURFACE " << fnSurface.name().asChar() 
			<< "\n\tID "		<< iNode
			<< "\n\tParentID "	<< iParent
			<< "\n\tmaterial "  << iMat
			<< "\n\tDegreeU "	<< fnSurface.degreeU()
			<< "\n\tDegreeV "	<< fnSurface.degreeV()
			<< "\n\tTypeU "		<< formu
			<< "\n\tTypeV "		<< formv
			<< "\n\tKnotsU "	<< (fnSurface.numKnotsInU()+2) << "\n\t\t";
			
		if( abs( static_cast<double>(fnSurface.knotInU(0) - fnSurface.knotInU(1)) ) < 0.01f )
			ofs << (float)fnSurface.knotInU(0) << " ";
		else
			ofs << (float)( fnSurface.knotInU( 0 ) - 1 ) << " ";
		
		for( i=0;i<fnSurface.numKnotsInU(); ++i )
			ofs << (float)fnSurface.knotInU(i) << " ";		
		
		if( abs( static_cast<double>(fnSurface.knotInU(0) - fnSurface.knotInU(1)) ) < 0.01f )
			ofs << (float)( fnSurface.knotInU( fnSurface.numKnotsInU()-1 ) ) << "\n";
		else
			ofs << (float)(fnSurface.knotInU( fnSurface.numKnotsInU()-1 )+1) << "\n";
		
		
		ofs << "\tKnotsV " << (fnSurface.numKnotsInV()+2) << "\n\t\t";
			
		if( abs( static_cast<double>(fnSurface.knotInV(0) - fnSurface.knotInV(1))) < 0.01f )
			ofs << (float)fnSurface.knotInV(0) << " ";
		else
			ofs << (float)( fnSurface.knotInV( 0 ) - 1 ) << " ";
		
		for( i=0;i<fnSurface.numKnotsInV(); ++i )
			ofs << (float)fnSurface.knotInV(i) << " ";		
		
		if( abs( static_cast<double>(fnSurface.knotInV(0) - fnSurface.knotInV(1)) ) < 0.01f )
			ofs << (float)( fnSurface.knotInV( fnSurface.numKnotsInV()-1 ) ) << "\n";
		else
			ofs << (float)(fnSurface.knotInV( fnSurface.numKnotsInV()-1 )+1) << "\n";
		
		
		ofs << "\tCVS " << fnSurface.numCVsInU() << " " << fnSurface.numCVsInV() << "\n";
		MPointArray pnts;
		fnSurface.getCVs(pnts);
		
		for(i=0;i<pnts.length();++i)
		{
			ofs << "\n\t\t" << (float)pnts[i].x << " " << (float)pnts[i].y << " " << (float)pnts[i].z;
		}
		
		ofs << "\n";
	}
}



