//////////////////////////////////////////////////////////////////////////////////////////////
/// \file	SkinCluster.cpp
/// \author	Robert Bateman
/// \brief	If you want to use skinning within your game, you will need to get hold of 
///			the skin cluster node created within maya. This holds all of the skin weights
/// 		and references to the influencing transforms (joints, ik handles, influence 
/// 		objects etc).
/// \note	
/// 		This sourcefile is provided for free use for non-commercial purposes only. The
/// 		Software is provided 'as is' and no responsibility can be taken for any damages
/// 		that may be caused while using it. If you make use of this software within 
/// 		your code then it would be nice to have a small credit somewhere in your
/// 		finished product for my hard work :(
/// 		An e-mail saying you found it useful and possibly even a link to your product
/// 		would be really nice too ;)
///
///			Slip me a couple of quid and i might be kind enough to offer assistance in
/// 		sorting out your exporters ;) 
///
///				(C) 2004 Rob 'give me a job offer' Bateman 

#ifdef WIN32
	#ifndef NT_APP
		#define NT_APP
	#endif

	// crappy VC6 template warnings
	#pragma warning(disable:4786)
#endif



#include "SkinCluster.h"
#include "Bone.h"

#include <maya/MFnSkinCluster.h>
#include <maya/MFnDependencyNode.h>
#include <maya/MFloatArray.h>
#include <maya/MDagPath.h>
#include <maya/MDagPathArray.h>
#include <maya/MItGeometry.h>

/// \brief	Holds a reference to a skin cluster in the file.
/// 
struct SSkinRef 
{
	SSkinRef() {}
	SSkinRef(const SSkinRef& rhs) {
		obj = rhs.obj;
		JointIDs = rhs.JointIDs;
	}
	const SSkinRef& operator = (const SSkinRef& rhs)
	{
		obj = rhs.obj;
		JointIDs = rhs.JointIDs;
		return *this;
	}

	/// the MObject handle to the skin cluster
	MObject obj;

	/// an array of the bone indices used for this skin cluster....
	std::vector<unsigned int> JointIDs;
};

std::vector<SSkinRef> g_aSkinClusterRefs;



/// \brief	This function takes an MObject for a skin cluster and then ensures that all bones
///			required for the cluster are going to be exported to the file with it....
// 
void AddSkinClusterRef(const MObject& obj) {
	MFnSkinCluster fnSkin(obj);
	MDagPathArray infs;
	MStatus stat;
	unsigned int nInfs = fnSkin.influenceObjects(infs, &stat);


	SSkinRef skin;
	skin.obj = obj;

	for(int i=0;i!=infs.length();++i)
	{
		skin.JointIDs.push_back( BuildJointRef(infs[i].node(),true) );
	}

	g_aSkinClusterRefs.push_back(skin);
}

/////////////////////////////////////////////////////////////////////////////////////////
/// this function outputs the specified mesh list
/// 
void OutputSkinClusters(std::ostream& ofs)
{
	MStatus stat;
	std::vector<SSkinRef>::iterator it = g_aSkinClusterRefs.begin();

	for( ; it != g_aSkinClusterRefs.end(); ++it )
	{
		MFnSkinCluster fnSkin(it->obj);

		// loop through the geometries affected by this cluster
		//
		unsigned int nGeoms = fnSkin.numOutputConnections();
		for (size_t ii = 0; ii < nGeoms; ++ii)
		{
			unsigned int index = fnSkin.indexForOutputConnection(static_cast<unsigned int>(ii),&stat);

			// get the dag path of the ii'th geometry
			//
			MDagPath skinPath;
			stat = fnSkin.getPathAtIndex(index,skinPath);

			// iterate through the components of this geometry
			//
			MItGeometry gIter(skinPath);

			// print out the path name of the skin, vertexCount & influenceCount
			//
			MString name = fnSkin.name();


			ofs << "SKIN " << fnSkin.name().asChar() 
				<< "\n\tNumInfluences " << static_cast<unsigned int>(it->JointIDs.size())
				<< "\n\t\t";

			std::vector<unsigned int>::iterator itJoint = it->JointIDs.begin();

			for( ; itJoint != it->JointIDs.end(); ++itJoint )
			{
				ofs << *itJoint << " ";
			}

			ofs << "\n\tMesh " << GetObjectID(skinPath.partialPathName().asChar())
				<< "\n\tNumVertices " << gIter.count();

			// because this exporter is explicitly for a game, each vertex can only have a 
			// maximum of four weights.

			unsigned int i =0,k;
			unsigned int NumVerts = gIter.count();

			ofs << "\n\tWeights\n";
			for ( i=0; i != NumVerts; ++i )
			{
				MObject comp = gIter.component();

				// Get the weights for this vertex (one per influence object)
				//
				MFloatArray wts;

				unsigned int infCount;

				fnSkin.getWeights(skinPath,comp,wts,infCount);


				if (0 != infCount && gIter.index() == i && !gIter.isDone())
				{
					unsigned int numWeights=0;

					float OutputWeights[40]={1.0f,0};

					unsigned int OutputInfs[40]={0};


					// Output the weight data for this vertex
					//
					for (unsigned int jj = 0; jj != infCount ; ++jj )
					{
						if(wts[jj] > 0.001f)
						{
							if(numWeights!=0)
							{
								OutputWeights[0]		   -= wts[jj];
								OutputWeights[numWeights]	= wts[jj];
							}
							OutputInfs[numWeights]		= jj;
							++numWeights;
						}
					}

					ofs << "\t\t" << numWeights << "\t";

					for(k=0;k!=numWeights;++k)
					{
						ofs << " " << OutputInfs[k] << " " << OutputWeights[k];
					}

					gIter.next();

				} else {
					ofs << "\t\t0";
				}
		

				ofs << std::endl;
 
			}
		}
	}

}

