/////////////////////////////////////////////////////////////////////////////////////////
/// 
///	\file	Texture.h
/// \author	Rob Bateman			( robthebloke@hotmail.com )
/// \brief	This file defines a bunch of functions to grab and export texture data. The
///			DoTextures function should be called before reading back any material data.
///			This will write all textures to a single binary file. The textures are then
///			assigned a 1 based index ID number. This information is stored internally 
///			in a map that uses the texture node name (NOT THE FILENAME!) as the key
///			and the ID as the value. 
/// 
///			This then allows you to use the GetTextureID() function later on to find the
///			reference number of the texture file.
/// 
///			CleanupMyTextures() should be called after the file has been exported. If not,
///			residual data may cause problems (not an issue just yet....)
///
///
/// \note	
/// 		This sourcefile is provided for free use for non-commercial purposes only. The
/// 		Software is provided 'as is' and no responsibility can be taken for any damages
/// 		that may be caused while using it. If you make use of this software within 
/// 		your code then it would be nice to have a small credit somewhere in your
/// 		finished product for my hard work :(
/// 		An e-mail saying you found it useful and possibly even a link to your product
/// 		would be really nice too ;)
///
///			Slip me a couple of quid and i might be kind enough to offer assistance in
/// 		sorting out your exporters ;) 
///
///				(C) 2004 Rob 'give me a job offer' Bateman 
/////////////////////////////////////////////////////////////////////////////////////////

#ifdef WIN32
#pragma once
#endif

#ifndef TEXTURE_H_INCLUDED
#define TEXTURE_H_INCLUDED



/////////////////////////////////////////////////////////////////////////////////////////

#include <maya/MString.h>



/////////////////////////////////////////////////////////////////////////////////////////


/// \brief	This struct is used as a 2byte identifier just before an image
/// \note	sizeof(ImageHeader) == 2
/// 
struct ImageHeader
{
	/// the width of the image as the power of two
	unsigned char m_WidthPower:4;
	/// the height of the image as a power of two
	unsigned char m_HeightPower:4;

	/// the bits per pixel
	/// 0 = 8bit
	/// 1 = 16bit
	/// 2 = 24bit
	/// 3 = 32bit
	/// 
	unsigned char m_Bpp:2;

	/// 6bits for a rainy day...
	unsigned char m_Unused:6;

	ImageHeader() : m_WidthPower(0),m_HeightPower(0),m_Bpp(0),m_Unused(0){

	}
};



/////////////////////////////////////////////////////////////////////////////////////////

/// \brief	This function outputs all of the textures in the scene into a single 
///			binary file. 
/// \param	outfile	-	the name of the binary texture file to write
/// 
void DoTextures(const char *const);


/// \brief	This function returns the texture ID of the texture node specified.
///			If the name specified does not exist it will return 0.
/// \param	fnn	-	the file node name. NOTE This is not the name of the texture
///			file, but the name of the fileTexture node in maya.
/// 
unsigned int GetTextureID(MString&);


/// \brief	This function cleans all dynamically allocated data (just the map infact)
/// 
void CleanupMyTextures();

#endif
