#include "BezierPatch.h"

XBezierPatch::XBezierPatch(void)
{
	m_MaxLOD = 20;
}

XBezierPatch::~XBezierPatch(void)
{
}

//-------------------------------------------------------------------
/// \brief	evaluates the surface using the provided surface points
/// \param	UseVerts	-	the vertices to use.
/// 
void XBezierPatch::Evaluate(XVertList* UseVerts) 
{
	// tesselate surface in one direction to create a set of temporary
	// curves that can be used for the second direction
	GenerateTemporaryVertices(UseVerts);

	// generate the final set of vertices
	GenerateVertices();

	// now generate the normals (i do this innacurately, though its
	// much faster - performance vs quality again....)
	GenerateNormals();
}

//-------------------------------------------------------------------
/// \brief	This is overidden to update the level of detail on the 
///			parametric shape. This will pre-calculate anything that
///			only changes when the LOD changes, this includes :
/// 
///			texture coords
///			vertex array indices
///			the blending factors for U and V
/// 
void XBezierPatch::OnSetLod()
{
	// num blend funcs = m_NumCvsU * GetLOD()
	m_BlendFuncs.resize(4 * GetLOD());

	// resize the temp points array
	m_TempPoints.resize(4 * GetLOD());

	// delete old vertex array data
	delete [] m_VertexArray;
	delete [] m_NormalArray;
	delete [] m_UvArray;
	delete [] m_IndexArray;

	// resize vertex array data
	m_VertexArray = new float[ 3 * GetLOD() * GetLOD() ];
	m_NormalArray = new float[ 3 * GetLOD() * GetLOD() ];
	m_UvArray     = new float[ 2 * GetLOD() * GetLOD() ];
	m_IndexArray  = new unsigned int[ 2 * GetLOD() * (GetLOD()-1) ];

	// generate data that does not change per frame
	GenerateTexCoords();
	GenerateIndices();

	std::vector<float>::iterator it = m_BlendFuncs.begin() ;

	register float t    = 0.0f;
	register float invt = 1.0f;
	float tinc = 1.0f/(GetLOD()-1);

	for ( ; it != m_BlendFuncs.end(); t += tinc, invt -= tinc )
	{
		// use temps for useful values
		register float t_sqr  = t*t;
		register float t_cube = t*t_sqr;
		register float invt_sqr  = invt*invt;
		register float invt_cube = invt*invt_sqr;

		*it = invt_cube;
		++it;
		*it = invt_sqr * t     * 3.0f;
		++it;
		*it = invt     * t_sqr * 3.0f;
		++it;
		*it = t_cube;
		++it;
	}
}

//------------------------------------------------------------------- Hard working functions !

//-------------------------------------------------------------------
///	\brief  This function evaluates the vertex positions
///	
void XBezierPatch::GenerateVertices()
{
	// wipe all vertex values to zero...
	memset(m_VertexArray,0,sizeof(float)*3*GetLOD()*GetLOD());

	// pointer used to step through output vertex array
	float* pOut = m_VertexArray;

	// used to keep track of the start point of the current curve
	// we are tesselating.
	std::vector<XVector>::iterator pPoint = m_TempPoints.begin();

	// surface is LOD * LOD output vertices...	therefore expect
	// a for loop....
	// 
	for( int i=0; i != GetLOD(); ++i,++pPoint )
	{ 
		// get an iterator to the blending funcs in U direction
		std::vector<float>::iterator itb = m_BlendFuncs.begin();

		// for each curve generate the output point 
		for( int j=0; j!=GetLOD(); ++j )
		{
			// get an iterator for the start of the line
			std::vector<XVector>::iterator pvert = pPoint;

			// loop through each vertex in this curve
			for( int k=0;k!=4; ++k, ++itb, pvert += GetLOD() )
			{
				// blending factor for this point
				float B = *itb;

				// comp quicker than vector mult, going to be more blending
				// funcs of zero than anything else, so may as well test for
				// the blending factor having no effect before calculating 
				// anything
				// 
				if( B > 0 ) 
				{
					// get quick pointer to output data
					float *p = pOut;

					// add the weighted sum of the points
					*p += B * pvert->x;		++p;
					*p += B * pvert->y;		++p;
					*p += B * pvert->z;
				}
			}

			// move on a bit in the output array
			pOut += 3;
		}		
	}

}


//-------------------------------------------------------------------
///	\brief  This function evaluates the vertex positions for the 
///			intermediate surface curves.
///	\param	UseVerts	-	an array of vertices to use for the 
///							control vertices.
/// 
void XBezierPatch::GenerateTemporaryVertices(XVertList* UseVerts)
{
	// get an iterator to the output data...
	std::vector<XVector>::iterator it = m_TempPoints.begin();

	// get an iterator to the input points
	XVertList::const_iterator itv = UseVerts->begin();

	// loop through each curve in the U direction
	for( int i=0;i!=4; ++i,++itv )
	{
		// get an iterator to the blending funcs in U direction
		std::vector<float>::iterator itb = m_BlendFuncs.begin();

		// generate each output point on curve 
		for( int j=0; j!=GetLOD(); ++j, ++it )
		{
			// temp reference to the control points
			XVertList::const_iterator itt = itv;

			// start with zero point..
			it->Set(0,0,0);

			for( int k=0; k != 4; ++k, itt += 4, ++itb )
			{
				//	*itt	=	the control point
				//	*itb	=	the blending factor for this point
				//	*it		=	output temporary vertex
				//
				// comp quicker than vector mult, going to be more blending
				// funcs of zero than anything else, so may as well test for
				// the blending factor having no effect before calculating 
				// anything
				// 
				if( (*itb) > 0 )
					*it +=  (*itb) * XVector(itt->x,itt->y,itt->z);
			}
		}
	}
}


std::istream &operator>>(std::istream& ifs,XBezierPatch& nc) {
	char buffer[64];
	ifs >> buffer; // name
	ifs >> buffer; // "id"
	ifs >> nc.m_ID;
	ifs >> buffer; // "ParentID"
	ifs >> nc.m_ParentID;
	ifs >> buffer;// "material"
	ifs >> nc.m_Material;

	nc.m_CVs.resize(16);
	for( int i=0;i!=16;++i)
	{
		ifs >> nc.m_CVs[i];
	}

	nc.SetLOD(20);
	return ifs;
}
