///-----------------------------------------------------------------------------
/// 
/// \file	Bone.h
/// \author	Rob Bateman
/// \brief	This file defines the class XBone which defines a bone/transform 
/// 		hierachy
/// \note	
/// 		This sourcefile is provided for free use for non-commercial purposes only. The
/// 		Software is provided 'as is' and no responsibility can be taken for any damages
/// 		that may be caused while using it. If you make use of this software within 
/// 		your code then it would be nice to have a small credit somewhere in your
/// 		finished product for my hard work :(
/// 		An e-mail saying you found it useful and possibly even a link to your product
/// 		would be really nice too ;)
///
///			Slip me a couple of quid and i might be kind enough to offer assistance in
/// 		sorting out your exporters ;) 
///
///				(C) 2004 Rob 'give me a job offer' Bateman 
/// 
///-----------------------------------------------------------------------------

#include <stdlib.h>
#ifdef WIN32
#define WIN32_LEAN_AND_MEAN
#include <Windows.h>
#endif
#include <GL/glut.h>
#include <iostream>
#include "Bone.h"


XBone::XBone() 
	: XFn(),
	  m_ID(0),
	  m_ParentID(0),
	  m_ChildJoints(),
	  m_ChildObjects(),
	  m_Xform(),
	  m_RotationOrient(),
	  m_JointOrient(),
	  m_BasePose(),
	  m_WorldSpace(),
	  m_LocalSpace()
{
}

XBone::XBone(const XBone& B) 
	: XFn(B),
	  m_ID(B.m_ID),
	  m_ParentID(B.m_ParentID),
	  m_ChildJoints(B.m_ChildJoints),
	  m_ChildObjects(B.m_ChildObjects),
	  m_Xform(B.m_Xform),
	  m_RotationOrient(B.m_RotationOrient),
	  m_JointOrient(B.m_JointOrient),
	  m_BasePose(B.m_BasePose),
	  m_WorldSpace(B.m_WorldSpace),
	  m_LocalSpace(B.m_LocalSpace)
{
}

XBone::~XBone()
{
}

std::istream& operator >> (std::istream& ifs,XBone &b)
{
	char buffer[64];
	int temp;
	ifs >> buffer; // joint name
	b.name = buffer;
	ifs	>> buffer // "id"
		>> b.m_ID 
		>> buffer // "parent"
		>> b.m_ParentID
		>> buffer // "children"
		>> temp;


	{

		for( int ttt = 0; ttt != temp; ++ttt )
		{
			int test;
			ifs >> test;
			if (test != b.m_ID)
			{
				b.m_ChildJoints.push_back(test);
			}
		}
	}

	ifs >> buffer // "ParentedNodes"
		>> temp;

	b.m_ChildObjects.resize(temp);

	{
		XBone::ChildIt it = b.m_ChildObjects.begin();

		for( ;it != b.m_ChildObjects.end(); ++it )
		{
			ifs >> *it;
		}
	}

	ifs >> buffer // "ParentedNodes"
		>> b.m_Xform.m_Scale;
	ifs >> buffer // "translation"
		>> b.m_Xform.m_Translation;
	ifs	>> buffer // "RotationOrient"
		>> b.m_RotationOrient;
	ifs	>> buffer // "Rotation"
		>> b.m_Xform.m_Rotation;
	ifs	>> buffer // "JointOrient"
		>> b.m_JointOrient;
	ifs >> buffer // "BaseMatrix"
		>> b.m_BasePose;


	
	return ifs;
}


void XBone::DrawBones( const XMatrix& ParentX,
					   std::vector< XBone*   > &Bones , 
					   std::vector< XTransform >& xforms,
					   std::vector< XObject* > &Objects,
					   std::vector< XMaterial*> &Materials  ) 
{

	glPushMatrix();

		XMatrix OrientationMatrix,
				RotationMatrix;

		// get a reference to the transform for this bone
		XTransform& xform = xforms[m_ID-1];

		// if drawing the bones, make sure we ignore the world to root bones
		if( 0 != m_ParentID )
		{
			// to get access to options
			XOptions Options;

			// if drawing joints
			if( Options.DrawJoints() ) 
			{
				glColor3f(0.4f,0.8f,0.8f);

				glDisable(GL_LIGHTING);
				glDisable(GL_TEXTURE_2D);

				glBegin(GL_LINES);
					glVertex3f(0.001f,0.001f,0.001f);		
					glVertex3fv( xform.m_Translation.data );
				glEnd();

				glEnable(GL_LIGHTING);
				glEnable(GL_TEXTURE_2D);
			}
		}

		// convert the orientation quaternion to a matrix
		/// \todo	err? rob? couldn't you just store an orientation
		///			in the joint to avoid having to convert the matrix each
		///			time??
		m_JointOrient.AsMatrix(OrientationMatrix.data);
		
		// add the translation into the matrix
		OrientationMatrix.data[12] = xform.m_Translation.x;
		OrientationMatrix.data[13] = xform.m_Translation.y;
		OrientationMatrix.data[14] = xform.m_Translation.z;

		
		// apply orientation adjustment
		glMultMatrixf(OrientationMatrix.data);

		// get the joints rotation as a matrix
		xform.m_Rotation.AsMatrix(RotationMatrix.data);
		
		// apply rotation
		glMultMatrixf(RotationMatrix.data);

		// local space xform = OrientationMatrix * RotationMatrix
		m_LocalSpace = OrientationMatrix * RotationMatrix;

		// apply scale to local space matrix
		m_LocalSpace.Scale(xform.m_Scale);

		// world space = ParentMatrix * LocalSpaceMatrix
		m_WorldSpace = ParentX * m_LocalSpace;

		// apply scale to openGL transform
		glScalef(xform.m_Scale.x,xform.m_Scale.y,xform.m_Scale.z);

		// loop through and draw all surface shapes that are parented under this bone
		ChildIt it;
		for( it = m_ChildObjects.begin(); it != m_ChildObjects.end(); ++it )
		{
			std::vector< XObject* >::iterator itO = Objects.begin();
			for( ; itO != Objects.end(); ++itO )
			{
				if( (*itO)->GetID() == *it )
				{
					(*itO)->gl(Materials);
				}
			}
		}

		//	loop through and recurse down each child.
		for( it = m_ChildJoints.begin(); it != m_ChildJoints.end(); ++it )
		{
			std::vector< XBone* >::iterator itB = Bones.begin();
			for( ; itB != Bones.end(); ++itB )
			{
				if( (*itB)->m_ID == *it )
				{
					(*itB)->DrawBones(  m_WorldSpace,
										Bones,
										xforms,
										Objects,
										Materials );
				}
			}
		}

	glPopMatrix();
}

bool XBone::RemoveObject(unsigned int child) {
	ChildIt it = m_ChildObjects.begin();
	for (;it != m_ChildObjects.end(); ++it )
	{
		if (*it == child)
		{
			m_ChildObjects.erase(it);
			return true;
		}
	}
	return false;
}
bool XBone::AddObject(unsigned int child) {
	ChildIt it = m_ChildObjects.begin();
	for (;it != m_ChildObjects.end(); ++it ) {
		if (*it == child) {
			return false;
		}
	}
	m_ChildObjects.push_back(child);
	return true;
}
