///-----------------------------------------------------------------------------
/// 
/// \file	Bone.h
/// \author	Rob Bateman
/// \brief	This file defines the class XBone which defines a bone/transform 
/// 		hierachy
/// \note	
/// 		This sourcefile is provided for free use for non-commercial purposes only. The
/// 		Software is provided 'as is' and no responsibility can be taken for any damages
/// 		that may be caused while using it. If you make use of this software within 
/// 		your code then it would be nice to have a small credit somewhere in your
/// 		finished product for my hard work :(
/// 		An e-mail saying you found it useful and possibly even a link to your product
/// 		would be really nice too ;)
///
///			Slip me a couple of quid and i might be kind enough to offer assistance in
/// 		sorting out your exporters ;) 
///
///				(C) 2004 Rob 'give me a job offer' Bateman 
/// 
///-----------------------------------------------------------------------------

#ifdef WIN32
#pragma once
#endif

#ifndef __X__BONE__H__
#define __X__BONE__H__
 

#include <vector>
#include <iostream>
#include <math.h>
#include <Vector.h>
#include <Matrix.h>
#include <Transform.h>
#include <Plane.h>

#include "Object.h"

#ifndef OUT
#define OUT
#endif

class XBone
	: public XFn
{

	friend class XBoneAnimation;
	friend std::istream& operator >> (std::istream& ifs,XBone &q);

public:
	typedef std::vector<unsigned int> ChildArray;
	typedef ChildArray::iterator ChildIt;

	XBone();

	~XBone();

	XBone(const XBone& B) ;

	//-----------------------------------------------------------------------
	/// \brief	Returns the ID of the bone
	///	\return	This transforms ID
	/// 
	unsigned int GetID() const {
		return m_ID;
	}

	//-----------------------------------------------------------------------
	/// \brief	returns the ID of the parent transform (0 if a root joint)
	///	\return	The parent ID
	///
	unsigned int GetParentID() const {
		return m_ParentID;
	}

	//-----------------------------------------------------------------------
	/// gets the type of the object
	virtual XFn::Type GetType() const {
		return XFn::Bone;
	}

	//-----------------------------------------------------------------------
	///	\brief  This function draws the bones and all surfaces parented 
	///			under them. It recurses through the hierarchy and builds
	///			up the world/local & inverse transform matrices. 
	///	\param  M			-	the world space of the parent transform
	/// \param	Bones		-	an array of all the bones in the model
	/// \param	Transforms	-	The animated transforms to use for the 
	///							bone rotations/translations & scales
	/// \param	Materials	-	an array of materials for the surfaces
	/// \param	Q			-	the parents rotation. Orientation thing...
	///
	void DrawBones( const XMatrix& M ,
					std::vector< XBone* >&,
					std::vector< XTransform >&,
		            std::vector< XObject* >&,
					std::vector< XMaterial*>& ); 

	///-------------------------------------------------------------------------
	/// \brief	This function returns the last calculated bone rotation
	/// \return	the rotation value
	/// 
	inline const XQuaternion& Rotation() const {
		return m_Xform.Rotation();
	}

	///-------------------------------------------------------------------------
	/// \brief	This function returns the last calculated bone translation
	/// \return	the translation value
	/// 
	inline const XVector& Translation() const {
		return m_Xform.Translation();
	}

	///-------------------------------------------------------------------------
	/// \brief	This function returns the last calculated bone scale
	/// \return	the scale value
	/// 
	inline const XVector& Scale() const {
		return m_Xform.Scale();
	}

	//-------------------------------------------------------------------
	///	\brief  returns the local space matrix for this bone transform
	///	\return the local space transform matrix
	///
	inline const XMatrix& World() const {
		return m_WorldSpace;
	}
	//-------------------------------------------------------------------
	///	\brief  returns the local space matrix for this bone transform
	///	\return the local space transform matrix
	///
	inline const XMatrix& Local() const {
		return m_LocalSpace;
	}
	//-------------------------------------------------------------------
	///	\brief  returns the local space matrix for this bone transform
	///	\return the local space transform matrix
	///
	inline const XMatrix& Base() const {
		return m_BasePose;
	}

	bool RemoveObject(unsigned int child);
	bool AddObject(unsigned int child);
	
//protected:

	/// the ID number of this bone
	unsigned int m_ID;

	/// the ID number of the parent bone
	unsigned int m_ParentID;

	/// the reference numbers to the child joints
	ChildArray m_ChildJoints;

	/// the reference numbers to any parented geometry/lights etc
	ChildArray m_ChildObjects;

	//////////////////////////////////////////////////////////////////////////
	/// 
	/// 
public:
	/// the translation / scale and rotation of the joint expressed as a transform
	XTransform	m_Xform;

	/// the joints rotation as a quaternion. 
	XQuaternion m_RotationOrient;

	/// the joints rotation as a quaternion. 
	XQuaternion m_JointOrient;
	
	/// the base pose of the bone
	XMatrix m_BasePose;

	/// the base pose of the bone
	XMatrix m_WorldSpace;

	/// the base pose of the bone
	XMatrix m_LocalSpace;

	/// the name of the joint.
	std::string name;

	//////////////////////////////////////////////////////////////////////////
};
typedef XBone::ChildArray ChildArray;
typedef XBone::ChildIt ChildIt;

typedef std::vector<XBone*> XBoneList;

#endif

