///-----------------------------------------------------------------------------
/// 
/// \file	BoneAnimation.h
/// \author	Rob Bateman
/// \brief	This file defines the class XBoneAnimation which holds the animation 
///			data for one bone / animation cycle. Later on I will replace this 
///			with an animation mixer to blend different animation cycles 
///			together and also allows for lower/upper body seperation.
/// \note	
/// 		This sourcefile is provided for free use for non-commercial purposes only. The
/// 		Software is provided 'as is' and no responsibility can be taken for any damages
/// 		that may be caused while using it. If you make use of this software within 
/// 		your code then it would be nice to have a small credit somewhere in your
/// 		finished product for my hard work :(
/// 		An e-mail saying you found it useful and possibly even a link to your product
/// 		would be really nice too ;)
///
///			Slip me a couple of quid and i might be kind enough to offer assistance in
/// 		sorting out your exporters ;) 
///
///				(C) 2004 Rob 'give me a job offer' Bateman 
/// 
///-----------------------------------------------------------------------------

#ifdef WIN32
#pragma once
#endif

#ifndef BONE_ANIMATION_H_INCLUDED_
#define BONE_ANIMATION_H_INCLUDED_

#include "Bone.h"


///-----------------------------------------------------------------------------
/// \brief	Not using this yet, but I will use it to replace the parameters
///			in each XBoneAnimation class. ie, one instance of this class will
///			be shared between a number of animation data chunks. This should
///			reduce the amount of data used by a little bit.... 
/// 
class XTimeInfo
{
public:

	/// the constructor
	XTimeInfo()
	{
		memset(this,0,sizeof(XTimeInfo));
	}

	///-------------------------------------------------------------------------
	/// \brief	this function is used to initialised the time control class.
	///	\param	start		-	the start time (in seconds) for the animation.
	/// \param	end			-	the end time (in seconds) for the animation
	/// \param	numframes	-	the number of frames in the cycle
	/// 
	void Initialise(const float start,const float end,const unsigned int numframes)
	{
		m_iNumFrames = numframes;
		m_fStart     = start;
		m_fEnd       = end;
		m_fLength    = end-start;
		m_fFrameTime = m_fLength/numframes;
		m_fCurrTime  = start;
	}

	///-------------------------------------------------------------------------
	/// \brief	This function updates the time by the specified delta
	/// \param	dt	-	difference in time
	/// 
	void Update(const float dt) 
	{		
		m_fCurrTime += dt;

		while(m_fCurrTime>m_fEnd) 
		{
			m_fCurrTime -= m_fLength;
		}
	}

	///-------------------------------------------------------------------------
	/// \brief	this function returns the current frame number
	/// \return	the frame index
	/// 
	inline unsigned int GetFrameNum() const
	{
		return (unsigned int)( (float)(m_fCurrTime - m_fStart)/m_fFrameTime );
	}

	///-------------------------------------------------------------------------
	/// \brief	this function returns the interpolation t for the current frame.
	///			ie, interpolate frame n and n+1 by frameT
	/// 
	inline const float GetFrameT() const
	{
		return  fmodf( m_fCurrTime - m_fStart,m_fFrameTime ) / m_fFrameTime;
	}

	
	///-------------------------------------------------------------------------
	/// \brief	This function returns the t value over the whole animation length
	/// 
	inline const float GetInterpolatingT() const
	{
		return  (m_fCurrTime-m_fStart)/m_fLength;
	}

protected:

	/// the number of frames in the cycle
	unsigned int				m_iNumFrames;

	/// the time in seconds between each frame in the cycle.
	float						m_fFrameTime;

	/// the start time of the animation. 
	float						m_fStart;

	/// the end time of the animation
	float						m_fEnd;

	/// the length (in seconds) of the animation
	float						m_fLength;

	/// the current time in seconds
	float						m_fCurrTime;

};


///-----------------------------------------------------------------------------
/// \brief	This class holds the animation data for one bone/animation cycle.
///			At the moment this holds rotation,scale and translation keys. Later on 
///			I will show you how to reduce the amount of keyframes to reduce
///			the data storage requirements and also reduce the amount of calculation
///			required.
/// 
class XBoneAnimation
{
	/// stream insertion operator
	friend std::istream& operator >> (std::istream& ifs,XBoneAnimation &ba);

public:

	/// ctor
	XBoneAnimation();
	
	/// dtor
	~XBoneAnimation();

	///-------------------------------------------------------------------------
	/// \brief	this func updates the current anim cycle
	/// \param	bone	-	the bone to update
	/// 
	void Update(OUT XTransform& xform,const XTimeInfo& time);

	void SetSize(unsigned int sz) {
		m_Keys.resize(sz);
	}

protected:

	/// the transform keyframes
	std::vector< XTransform >	m_Keys;
};



#endif

