//-------------------------------------------------------------------
/// \file	HermiteCurve.cpp
/// \author	Rob Bateman
/// \brief	Defines the loading and drawing functions for a bezier
///			curve.
//-------------------------------------------------------------------

#include "HermiteCurve.h"

#define TRY_DIFFERENT_TRAVERSAL 1

//-------------------------------------------------------------------
/// \brief	ctor
///
XHermiteCurve::XHermiteCurve() 
	: XParametricCurve()
{
	/// only ever want 4 points
	m_CVs.push_back( XVertex(0,0,0) );
	m_CVs.push_back( XVertex(0,1,0) );
	m_CVs.push_back( XVertex(1,1,0) );
	m_CVs.push_back( XVertex(1,0,0) );

	m_MaxLOD = 30;
}

//-------------------------------------------------------------------
/// \brief	dtor
/// 
XHermiteCurve::~XHermiteCurve()
{
}

//-------------------------------------------------------------------
/// \brief	This is overidden to update the level of detail on the 
///			parametric shape
/// 
void XHermiteCurve::OnSetLod()
{
	// delete old vertex array data
	delete [] m_VertexArray;

	// allocate new data
	m_VertexArray = new float[ 3 * GetLOD() ];

	// set the size for blending factor storage
	m_BlendingFunctions.resize( 4 * GetLOD() );

	// some temporary values - trying tom minimise calculation
	float tinc = 1.0f/(GetLOD()-1);

	// t and 1-t
	register float t    = 0.0f;

	// I'm slightly concerned by the way the data gets laid out in
	// this temporary array. Ideally I want to process all data arrays
	// in order so this top version simply writes the bledning functions
	// in a slightly different order... 
#if TRY_DIFFERENT_TRAVERSAL
	// loop through all blending functions for this LOD
	std::vector<float>::iterator itA = m_BlendingFunctions.begin() ;
	std::vector<float>::iterator itB = m_BlendingFunctions.begin() + GetLOD();
	std::vector<float>::iterator itC = m_BlendingFunctions.begin() + 2*GetLOD();
	std::vector<float>::iterator itD = m_BlendingFunctions.begin() + 3*GetLOD();

	std::vector<float>::iterator end = itB;

	for ( ; itA != end; t += tinc )
	{
		register float t_sqr  = t*t;
		register float t_sqr3 = 3*t_sqr;
		register float t_cube = t*t_sqr;
		register float t_cube2 = 2*t_cube;

		*itA =      t_cube2 -     t_sqr3 + 1;
		*itB =     -t_cube2 +     t_sqr3;
		*itC =       t_cube - 2.0f*t_sqr + t;
		*itD =       t_cube -      t_sqr;

		++itA;
		++itB;
		++itC;
		++itD;
	}
#else
	std::vector<float>::iterator it = m_BlendingFunctions.begin() ;
	register float t    = 0.0f;

	for ( ; it != m_BlendingFunctions.end(); t += tinc )
	{
		// use temps for useful values
		register float t_sqr  = t*t;
		register float t_sqr3 = 3*t_sqr;
		register float t_cube = t*t_sqr;
		register float t_cube2 = 2*t_cube;

		*it = t_cube2 - t_sqr3 + 1;
		++it;
		*it = -t_cube2 + t_sqr3;
		++it;
		*it = t_cube - 2.0f*t_sqr + t;
		++it;
		*it = t_cube -      t_sqr;
		++it;
	}
#endif
}

//-------------------------------------------------------------------
///	\brief  Evaluates the current surface positions and normals
///
void XHermiteCurve::Evaluate(XVertList* UseVerts)
{
	assert(m_VertexArray);

#if TRY_DIFFERENT_TRAVERSAL

	// set all output vertex values to zero initially
	memset(m_VertexArray,0,sizeof(float)*3*GetLOD());

	// loop through the blending functions
	std::vector<float>::iterator itblend = m_BlendingFunctions.begin();

	// 
	XVertIt it  = UseVerts->begin();
	for ( ; it != UseVerts->end(); ++it )
	{
		std::vector<float>::iterator itlastblend = itblend + GetLOD();
		float *pvert = m_VertexArray;

		for ( ; itblend != itlastblend; ++itblend )
		{
			float B = (*itblend);
			float *indat  = it->data;
			float *enddat = indat+3;

			for ( ; indat!=enddat; ++pvert, ++indat )
			{
				(*pvert) += B * (*indat);
			}
		}
	}
#else
	// loop through the blending functions
	std::vector<float>::iterator itblend = m_BlendingFunctions.begin();

	// pointer to output vertex data
	float *pvert = m_VertexArray;
	float *pend  = pvert + 3*GetLOD();

	for ( ; pvert != pend; )
	{
		XVector V(0,0,0);

		// 
		XVertIt it  = UseVerts->begin();
		for ( ; it != UseVerts->end(); ++it,++itblend )
		{
            V += (*it) * (*itblend);		
		}

		*pvert = V.x;
		++pvert;
		*pvert = V.y;
		++pvert;
		*pvert = V.z;
		++pvert;
	}

#endif
}


//-------------------------------------------------------------------
/// \brief	ifs	-	the input file stream
/// \param	nc	-	the bezier curve to read
std::istream &operator>>(std::istream& ifs,XHermiteCurve& nc) {
	char buffer[64];
	ifs >> buffer; // name
	ifs >> buffer; // "id"
	ifs >> nc.m_ID;
	ifs >> buffer; // "ParentID"
	ifs >> nc.m_ParentID;
	
	nc.m_CVs.resize(4);
	for( int i=0;i!=4;++i)
	{
		ifs >> nc.m_CVs[i];
	}

	nc.SetLOD(20);
	return ifs;
}
