///-----------------------------------------------------------------------------
/// 
/// \file	Material.h
/// \author	Rob Bateman
/// \brief	I've tried to make this loader a bit more generic. The material 
///			class is now able to store colour channels, ie a colour and a texture
///			for ambient, diffuse, specular etc. With basic openGL this isn't of 
///			much use because you can only use 1 texture 
///			(though 2 with mutlitexturing extension). This is mainly done for my
///			own purposes cos I use CG shaders so a bit more flexibility is nice.
/// 
///			At the moment, if a texture is assigned to the diffuse channel, then 
///			it is used as the surface texture.
/// \note	
/// 		This sourcefile is provided for free use for non-commercial purposes only. The
/// 		Software is provided 'as is' and no responsibility can be taken for any damages
/// 		that may be caused while using it. If you make use of this software within 
/// 		your code then it would be nice to have a small credit somewhere in your
/// 		finished product for my hard work :(
/// 		An e-mail saying you found it useful and possibly even a link to your product
/// 		would be really nice too ;)
///
///			Slip me a couple of quid and i might be kind enough to offer assistance in
/// 		sorting out your exporters ;) 
///
///				(C) 2004 Rob 'give me a job offer' Bateman 
/// 
///-----------------------------------------------------------------------------



#ifdef WIN32
#pragma once
#endif

#ifndef __XMATERIAL__H_
#define __XMATERIAL__H_

#include <vector>
#include <iostream>

#include "Object.h"

class XMaterial 
	: public XFn
{
public:
	
	XMaterial(void);
	~XMaterial(void);

	/// the function to bind the material
	void gl();
	

	virtual XFn::Type GetType() const {
		return XFn::Material;
	}
	

	/// a function to set the ambient colour
	inline void SetAmbient(const float r,
						   const float g,
						   const float b,
						   const float a=1.0f) {
		m_Ambient.SetColour(r,g,b,a);
	}

	/// a function to set the diffuse colour
	inline void SetDiffuse(const float r,
						   const float g,
						   const float b,
						   const float a=1.0f) {
		m_Diffuse.SetColour(r,g,b,a);
	}

	/// a function to set the specular colour
	inline void SetSpecular(const float r,
						    const float g,
						    const float b,
						    const float a=1.0f) {
		m_Specular.SetColour(r,g,b,a);
	}

	/// a function to set the emission colour
	inline void SetEmission(const float r,
						    const float g,
						    const float b,
						    const float a=1.0f) {
		m_Emission.SetColour(r,g,b,a);
	}

	/// a function to set the emission colour
	inline void SetTransparency(const float r,
						        const float g,
						        const float b,
						        const float a=1.0f) {
		m_Transparency.SetColour(r,g,b,a);
	}

	/// a function to set the shininess of the material
	inline void SetShininess(const float shine) {
		m_Shininess = shine;
	}

	/// a function to set the texture object for the material to use
	inline void SetTextureStartNum(unsigned int tex) {
		if(m_Ambient.m_TextureObject != 0 ) m_Ambient.m_TextureObject += (tex-1);
		if(m_Diffuse.m_TextureObject != 0 ) m_Diffuse.m_TextureObject += (tex-1);
		if(m_Specular.m_TextureObject != 0 ) m_Specular.m_TextureObject += (tex-1);
		if(m_Emission.m_TextureObject != 0 ) m_Emission.m_TextureObject += (tex-1);
		if(m_Transparency.m_TextureObject != 0 ) m_Transparency.m_TextureObject += (tex-1);
	}

protected:


	class XColorChannel
	{
	public:

		XColorChannel() : m_TextureObject(0) {
			m_Colour[0] = 0;
			m_Colour[1] = 0;
			m_Colour[2] = 0;
			m_Colour[3] = 0;
		}

		inline float Clamp(const float &f) const {
			if (f>1.0f) {
				return 1.0f;
			}	
			if (f<0.0f) {
				return 0.0f;
			}
			return f;
		}
	
		friend std::istream& operator >> (std::istream& ifs,XColorChannel& channel)
		{
			char buffer[256];

			ifs >> buffer // channel name
				>> channel.m_Colour[0] >> channel.m_Colour[1] >> channel.m_Colour[2] >> channel.m_Colour[3]
				>> buffer // Tex=
				>> channel.m_TextureObject;
				return ifs;
		}

		/// texture ID
		unsigned int m_TextureObject;

		/// the RGBA colour
		float m_Colour[4];


		/// a function to set the emission colour
		inline void SetColour(const float r,
							  const float g,
							  const float b,
							  const float a=1.0f) {
			m_Colour[0] = Clamp(r);
			m_Colour[1] = Clamp(g);
			m_Colour[2] = Clamp(b);
			m_Colour[3] = Clamp(a);
		}
	};

	/// the ambient material colour
	XColorChannel m_Ambient;

	/// the diffuse colour
	XColorChannel m_Diffuse;

	/// the specular colour
	XColorChannel m_Specular;

	/// the emission colour
	XColorChannel m_Emission;

	/// the transparency colour
	XColorChannel m_Transparency;

	/// the shininess
	float m_Shininess;

	/// as an optimisation, we will track which material is currently in use. This will 
	/// try to minimise the state changes within OpenGL
	static XMaterial* ms_CurrentMaterial;
};


/// a list of material types
typedef std::vector< XMaterial* > XMaterialList;

/// an iterator type for the material list
typedef XMaterialList::iterator XItMaterial;

/// the phong material type
class XPhong : public XMaterial
{
public:
	friend std::istream& operator >> (std::istream& ifs,XPhong& ph);

	virtual XFn::Type GetType() const {
		return XFn::Phong;
	}

};

/// the lambert material type
class XLambert : public XMaterial
{
public:
	friend std::istream& operator >> (std::istream& ifs,XLambert& ph);

	virtual XFn::Type GetType() const {
		return XFn::Lambert;
	}
};

/// the blinn material type
class XBlinn : public XMaterial
{
public:
	friend std::istream& operator >> (std::istream& ifs,XBlinn& ph);

	virtual XFn::Type GetType() const {
		return XFn::Blinn;
	}
};


#endif

