///-----------------------------------------------------------------------------
/// 
/// \file	Mesh.h
/// \author	Rob Bateman
/// \brief	a simple file loader for a simple file... This loader basically 
///			uses streaming operators to read and write the file. The data is 
///			stored in a set os basic classes, XVertex, XNormal, XTexCoord and
///			XFace.
/// 
///			The XMesh class then holds the data within arrays, these array types 
///			are typedefs of std::vector<XVertex> etc.
/// \note	
/// 		This sourcefile is provided for free use for non-commercial purposes only. The
/// 		Software is provided 'as is' and no responsibility can be taken for any damages
/// 		that may be caused while using it. If you make use of this software within 
/// 		your code then it would be nice to have a small credit somewhere in your
/// 		finished product for my hard work :(
/// 		An e-mail saying you found it useful and possibly even a link to your product
/// 		would be really nice too ;)
///
///			Slip me a couple of quid and i might be kind enough to offer assistance in
/// 		sorting out your exporters ;) 
///
///				(C) 2004 Rob 'give me a job offer' Bateman 
/// 
///-----------------------------------------------------------------------------


//////////////////////////////////////////////////////////////////////////

#ifdef WIN32
#pragma once
#endif


#ifndef XMESH__H_
#define XMESH__H_

//////////////////////////////////////////////////////////////////////////
/// 
#ifdef WIN32
#define WIN32_LEAN_AND_MEAN
#include <Windows.h>
#endif
#include <GL/gl.h>

#include <vector>
#include <fstream>
#include <Vector.h>
#include <Transform.h>
#include <Plane.h>

#include "Material.h"
#include "Object.h"

/// this internal structure is used to package all indices
/// into 12bytes. This keeps the data access sequentual through
/// memory in the draw function
///
struct XPointIndex 
{
	/// streaming operator
	friend std::istream &operator>>(std::istream& ifs,XPointIndex& pi) ;

	/// i prefer not to initialise array indices because it often 
	/// highlights problems with the code that would be otherwise
	/// undetected. ie, the below indices are probably going to crash
	/// the program if they are not initialised. If I set these to 
	/// zero then I would never be aware of this problem....
	/// 
	XPointIndex() {}
	XPointIndex(const XPointIndex& v_) : t(v_.t),n(v_.n),v(v_.v) {}

	const XPointIndex& operator = (const XPointIndex& rhs) {
		n=rhs.n;
		t=rhs.t;
		v=rhs.v;
		return *this;
	}

	union {
		struct {
			/// the texture coordinate index
			unsigned int t;
		
			/// the normal index
			unsigned int n;
		
			/// the vertex index
			unsigned int v;
		};
		unsigned int data[3];
	};	
};


//////////////////////////////////////////////////////////////////////////
/// a class to hold a face
///
class XFace
{
	/// stream operator
	friend std::istream &operator>>(std::istream&,XFace&);

	/// the mesh class is a friend...
	friend class XMesh;


public:


	/// ctor - sets the pointer to NULL for that is good...
	XFace() 
		: m_Points(NULL),
		  m_NumPoints(0),
		  m_Material(0) {}

	/// dtor
	~XFace() {
		delete m_Points;
	}

	/// ctor - sets the pointer to NULL for that is good...
	XFace(const XFace& f) {
		m_NumPoints = f.m_NumPoints;
		m_Material  = f.m_Material;
		m_Points    = new XPointIndex[ m_NumPoints ];
		for(int i=0;i!=m_NumPoints;++i)
		{
			m_Points[i] = f.m_Points[i];
		}	
	}

	/// this function sets the number of indices in the face
	/// 
	void SetSize(unsigned int);

	const unsigned int& NumPoints() const {
		return m_NumPoints;
	}
	unsigned int GetVertexIndex(unsigned int ind) const {
		return m_Points[ind].v;
	}
	unsigned int GetNormalIndex(unsigned int ind) const {
		return m_Points[ind].v;
	}

protected:


	/// the number of points in the polygon
	unsigned int m_NumPoints;

	/// an array of point indices
	XPointIndex *m_Points;

	/// the material index for this face
	unsigned int m_Material;
};


//////////////////////////////////////////////////////////////////////////

// typedefs for readabilities sake
typedef std::vector<XFace> XFaceList;
typedef XFaceList::iterator XFaceIt;
typedef XFaceList::const_iterator XConstFaceIt;

//////////////////////////////////////////////////////////////////////////

class XMesh : public XObject
{
	friend std::istream &operator>>(std::istream&,XMesh&);

public:

	/// ctor
	XMesh() {};

	/// dtor
	~XMesh() {};

	
	/// this function draws a deformed version of the mesh
	/// \param	mats	-	the material array to use
	/// \param	UseVerts	-	the vertices to use
	/// \param	UseNorms	-	the normals to use
	virtual void gl(XMaterialList& mats,
					XVertList* UseVerts,
					XNormList* UseNorms);
	
	virtual XVertList* Points() {
		return &m_Vertices;
	}
	virtual XNormList* Normals() {
		return &m_Normals;
	}
	const XTexList& TexCoords() const {
		return m_TexCoords;
	}
	const XFaceList& Faces() const {
		return m_Faces;
	}
	const char* Name() const {
		return m_Name.c_str();
	}
	/// gets the type of the object
	virtual XFn::Type GetType() const {
		return XFn::Mesh;
	}
	
protected:

	/// an array of faces
	XFaceList m_Faces;

	/// an array of texture coords
	XTexList  m_TexCoords;
	
	/// an array of normals
	XNormList m_Normals;

	/// an array of vertices
	XVertList m_Vertices;

	/// the mesh name 
	std::string m_Name;

};


#endif

