///-----------------------------------------------------------------------------
/// 
/// \file	Model.cpp
/// \author	Rob Bateman
/// \brief	This file defines the XModel class. basicaly this class will load the 
///			contents of a mod file. This version merely holds lights and 
///			meshes in one place and is responsible for drawing them.
/// \note	
/// 		This sourcefile is provided for free use for non-commercial purposes only. The
/// 		Software is provided 'as is' and no responsibility can be taken for any damages
/// 		that may be caused while using it. If you make use of this software within 
/// 		your code then it would be nice to have a small credit somewhere in your
/// 		finished product for my hard work :(
/// 		An e-mail saying you found it useful and possibly even a link to your product
/// 		would be really nice too ;)
///
///			Slip me a couple of quid and i might be kind enough to offer assistance in
/// 		sorting out your exporters ;) 
///
///				(C) 2004 Rob 'give me a job offer' Bateman 
/// 
///-----------------------------------------------------------------------------

#ifdef WIN32
#define WIN32_LEAN_AND_MEAN
#include <Windows.h>
#endif
#include <GL/gl.h>

#include "Model.h"
#include "Mesh.h"
#include "Options.h"
#include "TextureFile.h"
#include "CollisionSphere.h"
#include "CollisionBox.h"
#include "CollisionCone.h"
#include "BezierCurve.h"
#include "BezierPatch.h"
#include "HermiteCurve.h"
#include "HermitePatch.h"
#include "Locator.h"

#include <assert.h>
#include <fstream>

#ifndef GL_EXT_separate_specular_color
#define GL_EXT_separate_specular_color
#define GL_LIGHT_MODEL_COLOR_CONTROL_EXT  0x81F8
#define GL_SINGLE_COLOR_EXT               0x81F9
#define GL_SEPARATE_SPECULAR_COLOR_EXT    0x81FA
#endif


XModel::XModel(void) : m_TextureObject(0),m_NumTextures(0)
{
}

XModel::~XModel(void)
{
	// delete tesxtures
	glDeleteTextures(m_NumTextures,&m_TextureObject);


	// delete all meshes
	{
		XIt it = m_Objects.begin();
		for( ; it != m_Objects.end(); ++it )
		{
			delete *it;
		}
	}
}

/// is the model textured?
bool XModel::IsTextured() const {
	return m_TextureObject != 0 ;
}

/// does the model contain lights?
bool XModel::IsLit() const {
	return m_Lights.size() != 0;
}

/// \brief	returns the number of animations that the model has loaded
/// \return	the number of anim cycles
/// 
unsigned int XModel::NumAnimations() const {
	return static_cast<unsigned int>(m_Anims.size());
}

/// \brief	This function sets the current animation to the specified id
/// \param	id	-	the id number of the animation to set as the current anim
/// \note	this is temporary until I get round to replacing it with blending 
/// 
void XModel::SetCurrentAnimation(unsigned int id) {

	// make sure id is valid
	assert(m_Anims.size() > id);

	// set the current animation pointer
	pCurrentAnimation = m_Anims[id];
}

XBone* XModel::GetBone( unsigned int id ) {
	XItBone it = m_Bones.begin();
	for (; it != m_Bones.end(); ++it )
	{
		if ((*it)->GetID() == id)
		{
			return *it;
		}
	}
	return NULL;
}

XObject* XModel::GetObject( unsigned int id ) {
	XIt it = m_Objects.begin();
	for (; it != m_Objects.end(); ++it )
	{
		if ((*it)->GetID() == id)
		{
			return *it;
		}
	}
	return NULL;
}

/// this Function enables all of the lights in the model
void XModel::EnableLights()
{
	// return false if no lights exist
	if (m_Lights.empty()) {
		return;
	}

	// enable lighting
	glEnable(GL_LIGHTING);

	// enable all lights
	XItLight it = m_Lights.begin();
	for(;it!=m_Lights.end();++it)
	{
		(*it)->Enable();
	}
}

/// this function disables all of the lights in the model
void XModel::DisableLights()
{
	// return if no lights exist
	if (m_Lights.empty()) {
		return;
	}
	
	// disable lighting
	glDisable(GL_LIGHTING);

	// disable each light
	XItLight it = m_Lights.begin();
	for(;it!=m_Lights.end();++it)
	{
		(*it)->Disable();
	}
}



/// a function to draw the geometry data in the model
void XModel::gl()
{

	XOptions Options;

	if(Options.SS())
		glLightModelf( GL_LIGHT_MODEL_COLOR_CONTROL_EXT, GL_SEPARATE_SPECULAR_COLOR_EXT );
	XMatrix M;

	glPushMatrix();

		EnableLights();
		glEnable(GL_TEXTURE_2D);
		glEnable(GL_LIGHT0);


		// if meshes are in wireframe
		if( Options.IsWireframe() )
			glPolygonMode(GL_FRONT_AND_BACK,GL_LINE);

		// if culling is enabled
		if( Options.IsCulled() )
			glEnable(GL_CULL_FACE);
		else
			glDisable(GL_CULL_FACE);

		// draw each mesh
		XItBone it = m_Bones.begin();
		for(;it!=m_Bones.end();++it)
		{
			// if a root node
			if((*it)->GetParentID() == 0) {

				// draw the bone hierarchy
				(*it)->DrawBones(M,m_Bones,m_BoneTransforms,m_Objects,m_Materials);
			}
		}

		// loop through objects such as skinned meshes and lattice deformers that are
		//	stored directly under the model (ie, they require world space)
		//
		ChildIt itc;
		for( itc = m_ParentedObjects.begin(); itc != m_ParentedObjects.end(); ++itc )
		{
			XObjectList::iterator itO = m_Objects.begin();
			for( ; itO != m_Objects.end(); ++itO )
			{

				XFn::Type type = (*itO)->GetType();
				switch(type) 
				{
				case XFn::BezierCurve:
				case XFn::BezierPatch:
				case XFn::HermiteCurve:
				case XFn::HermitePatch: 
				case XFn::NurbsCurve:
				case XFn::NurbsSurface: {
						XBone* B = GetBone( (*itO)->GetParentID() );
						XParametricBase::SetObjectPosition( *itO, B->World() * XVector() );
					}
					break;
				default:
					break;
				}


				if( (*itO)->GetID() == *itc )
				{
					(*itO)->gl(m_Materials);
				}
			}
		}


		glPolygonMode(GL_FRONT_AND_BACK,GL_FILL);

		DisableLights();

	glPopMatrix();

	if(Options.SS())
		glLightModelf( GL_LIGHT_MODEL_COLOR_CONTROL_EXT, GL_SINGLE_COLOR_EXT );
}

void XModel::Update(const float dt)
{

	if(pCurrentAnimation) {

		XItBone it = m_Bones.begin();
		pCurrentAnimation->GetTime().Update(dt);
		
		// loop through all of the bones in the model
		for( ;it != m_Bones.end(); ++it )
		{
			// if valid ptr, probably should use assert....
			if ( *it ) {

				assert( pCurrentAnimation && "Model has no animation cycle" );

				// get pointer to the bone animation from the anim cycle
				XBoneAnimation* pba = pCurrentAnimation->GetBoneAnim( (*it)->GetID() );

				// if anim found (use an assert?)
				if(pba)
				{
					// update the animation value
					pba->Update( m_BoneTransforms[(*it)->GetID()-1] , pCurrentAnimation->GetTime() );
				}
			}
		}
	}	
}


bool XModel::Parent(unsigned int child,unsigned int parent) {
	if (parent == 0)
	{
		UnParent(child);
		m_ParentedObjects.push_back(child);
		return true;
	}
	else {
		XBone* B = GetBone(parent);
		if (B)
		{
			XObject* Obj = GetObject(child);
			if (Obj)
			{
				XBone* PX = GetBone( Obj->GetParentID() );
				if(PX)
					PX->RemoveObject(child);
				B->AddObject(child);
				return true;
			}
		}
		std::cerr << "Parent: No bone found" << std::endl;
	}
	return false;
}


bool XModel::UnParent(unsigned int child) {
	XObject* Obj = GetObject(child);
	if (Obj)
	{
		unsigned int id = Obj->GetParentID();
		if (id != 0)
		{
			XBone* bone = GetBone(id);
			if(bone) 
			{
				std::cerr << "Removing from bone" << std::endl;
				return bone->RemoveObject(child);
			}
			else
			{
				std::cerr << "No bone found" << std::endl;
			}
		}
		else 
		{
			ChildIt it = m_ParentedObjects.begin();
			for ( ; it != m_ParentedObjects.end(); ++it )
			{
				if (*it == child)
				{
					std::cerr << "erasing from global parent" << std::endl;
					m_ParentedObjects.erase(it);
					return true;
				}
			}
			std::cerr << "Object not found bone" << std::endl;
		}
	}
	return false;
}



/// \brief	a function to load a file into the model
/// \param	filename	-	the name of the file to load
/// \return	true if OK
/// 
bool XModel::Load(const char* const filename)
{
	{
		// first of all, see if there is a texture file
		std::string texfile = filename;
		texfile.resize(texfile.size()-3);
		texfile += "tex";

		// load the texture file
		XTextureFile texLoader;
		if(texLoader.Load(texfile.c_str()))
		{
			// if OK, pass all textures to openGL
			if(texLoader.TurnIntoTextures(m_TextureObject))
			{
				// store number of textures
				m_NumTextures = texLoader.GetNumTextures();
			}
		}
	}

	char buffer[512];
	std::ifstream ifs(filename);

	if (ifs) {

		while (!ifs.eof()) {
			ifs >> buffer;
			if (ifs.eof()) {
				break;
			}

			std::cerr << buffer << std::endl;

			// check for the bone flag
			// 
			if (buffer[0] == 'A' &&	//------------------------------------ Anim Cycle
				buffer[1] == 'N' &&
				buffer[2] == 'I' &&
				buffer[3] == 'M' &&
				buffer[4] == 'C' &&	
				buffer[5] == 'Y' &&
				buffer[6] == 'C' &&
				buffer[7] == 'L' &&
				buffer[8] == 'E' )
			{
				// create anim cycle chunk
				XAnimCycle* ac = new XAnimCycle;
				assert (ac);

				// read anim cycle
				ifs >> *ac;

				// set new anim as the current animation
				pCurrentAnimation = ac;

				// add to cycle list
				m_Anims.push_back(ac);
				
			}
			else
			if (buffer[0] == 'B' )
			{
				if(	buffer[1] == 'O' &&	//-------------------------------- Collision Box
					buffer[2] == 'X' )
				{
					XCollisionBox* pskin = new XCollisionBox();
					assert (pskin);
					ifs >> *pskin;
					m_Objects.push_back(pskin);		

				}
				else
				if(	buffer[1] == 'O' &&	//-------------------------------- Bone
					buffer[2] == 'N' &&
					buffer[3] == 'E') 
				{
					// check for the animation flag
					// 
					if (buffer[4] == 'A' &&
						buffer[5] == 'N' &&
						buffer[6] == 'I' &&
						buffer[7] == 'M') 
					{
						assert( 0 && "SHOULD NOT GET HERE BoneAnim chunks are obsolete!!!");
					}
					// otherwise it is definately a bone....
					//
					else
					{
						// create bone chunk
						XBone* pb = new XBone;
						assert (pb);

						// read bone
						ifs >> *pb;

						// add to bone list
						m_Bones.push_back(pb);
					}
				}
				else
				if(	buffer[1] == 'E' &&
					buffer[2] == 'Z' &&
					buffer[3] == 'I' &&
					buffer[4] == 'E' &&
					buffer[5] == 'R' &&
					buffer[6] == '_' ) 
				{
					// check for bezier curve flag
					// 
					if (buffer[7]  == 'C' &&	//------------------------ Bezier Curve
						buffer[8]  == 'U' &&
						buffer[9]  == 'R' &&
						buffer[10] == 'V' &&
						buffer[11] == 'E') 
					{
						XBezierCurve* pc = new XBezierCurve;
						assert (pc);
						ifs >> *pc;
						m_Objects.push_back(pc);
					}
					else
					// check for bezier patch flag
					// 
					if (buffer[7]  == 'P' &&	//------------------------ Bezier Patch
						buffer[8]  == 'A' &&
						buffer[9]  == 'T' &&
						buffer[10] == 'C' &&
						buffer[11] == 'H') 
					{
						XBezierPatch* pc = new XBezierPatch;
						assert (pc);
						ifs >> *pc;
						m_Objects.push_back(pc);
					}
				}
				else
				if (buffer[1] == 'L' &&	//-------------------------------- Blinn
					buffer[2] == 'I' &&
					buffer[3] == 'N' &&
					buffer[4] == 'N' )
				{
					// create blinn material chunk
					XBlinn* pp = new XBlinn;
					assert (pp);

					// read from file
					ifs >> *pp;

					// if it uses a texture....
					if (m_TextureObject!=0) {

						pp->SetTextureStartNum( m_TextureObject );
					}

					// add to back of materials list
					m_Materials.push_back(pp);					
				}
			}
			else
			if (buffer[0] == 'L' )
			{
				if(	buffer[1] == 'I' &&	//-------------------------------- Light
					buffer[2] == 'G' &&
					buffer[3] == 'H' &&
					buffer[4] == 'T' ) 
				{
					XLight* pl = new XLight( static_cast<unsigned int>(m_Lights.size()));
					assert (pl);
					ifs >> *pl;
					m_Lights.push_back(pl);
					m_Objects.push_back(pl);
				}
				else
				if(	buffer[1] == 'A' &&	//-------------------------------- Lambert
					buffer[2] == 'M' &&
					buffer[3] == 'B' &&
					buffer[4] == 'E' &&
					buffer[5] == 'R' &&
					buffer[6] == 'T') 
				{
					XLambert* pp = new XLambert;
					assert (pp);
					ifs >> *pp;
					if (m_TextureObject!=0) {

						pp->SetTextureStartNum( m_TextureObject );
					}
					m_Materials.push_back(pp);
				}
				else
				if(	buffer[1] == 'O' &&	//-------------------------------- Locator
					buffer[2] == 'C' &&
					buffer[3] == 'A' &&
					buffer[4] == 'T' &&
					buffer[5] == 'O' &&
					buffer[6] == 'R') 
				{
					XLocator* pm = new XLocator;
					assert (pm);
					ifs >> *pm;
					m_Objects.push_back(pm);
				}
			}
			else
			if (buffer[0] == 'M' &&	//------------------------------------ Mesh
				buffer[1] == 'E' &&
				buffer[2] == 'S' &&
				buffer[3] == 'H') 
			{
				XMesh* pm = new XMesh;
				assert (pm);
				ifs >> *pm;
				m_Objects.push_back(pm);
			}
			else
			if (buffer[0] == 'C' &&	//------------------------------------ Collision Cone
				buffer[1] == 'O' &&
				buffer[2] == 'N' &&
				buffer[3] == 'E') 
			{
				XCollisionCone* pm = new XCollisionCone;
				assert (pm);
				ifs >> *pm;
				m_Objects.push_back(pm);
			}
			else
			if (buffer[0] == 'S' )
			{
				if(	buffer[1] == 'K' &&	//-------------------------------- Skin Cluster
					buffer[2] == 'I' &&
					buffer[3] == 'N') 
				{
					XSkinCluster* pskin =  new XSkinCluster( this );
					assert (pskin);
					ifs >> *pskin;
					m_Objects.push_back(pskin);
				}
				else
				if(	buffer[1] == 'P' &&	//-------------------------------- Collision Sphere
					buffer[2] == 'H' &&
					buffer[3] == 'E' &&
					buffer[4] == 'R' &&
					buffer[5] == 'E' ) 
				{
					XCollisionSphere* pskin = new XCollisionSphere();
					assert (pskin);
					ifs >> *pskin;
					m_Objects.push_back(pskin);					
				}
			}
			else
			if (buffer[0] == 'N' &&
				buffer[1] == 'U' &&
				buffer[2] == 'R' &&
				buffer[3] == 'B' &&
				buffer[4] == 'S' &&
				buffer[5] == '_') 
			{
				if(	buffer[6] == 'C' &&	//-------------------------------- Nurbs Curve
					buffer[7] == 'U' &&
					buffer[8] == 'R' &&
					buffer[9] == 'V' &&
					buffer[10] == 'E' ) 
				{
					XNurbsCurve* pc = new XNurbsCurve;
					assert (pc);
					ifs >> *pc;
					m_Objects.push_back(pc);
				}
				else
				if(	buffer[6] == 'S' &&	//-------------------------------- Nurbs Surface
					buffer[7] == 'U' &&
					buffer[8] == 'R' &&
					buffer[9] == 'F' &&
					buffer[10] == 'A'&&
					buffer[11] == 'C'&&
					buffer[12] == 'E' ) 
				{
					XNurbsSurface* ps = new XNurbsSurface;
					assert (ps);
					ifs >> *ps;
					m_Objects.push_back(ps);
				}
			}
			else
			if (buffer[0] == 'P' &&	//------------------------------------ Phong
				buffer[1] == 'H' &&
				buffer[2] == 'O' &&
				buffer[3] == 'N' &&
				buffer[4] == 'G' ) 
			{
				XPhong* pp = new XPhong;
				assert (pp);
				ifs >> *pp;
				if (m_TextureObject!=0) {

					pp->SetTextureStartNum( m_TextureObject );
				}
				m_Materials.push_back(pp);
			}
			else
			if(	buffer[0] == 'H' &&
				buffer[1] == 'E' &&
				buffer[2] == 'R' &&
				buffer[3] == 'M' &&
				buffer[4] == 'I' &&
				buffer[5] == 'T' &&
				buffer[6] == 'E' &&
				buffer[7] == '_' ) 
			{

				// check for bezier curve flag
				// 
				if (buffer[8]  == 'C' &&	//---------------------------- Hermite Curve
					buffer[9]  == 'U' &&
					buffer[10] == 'R' &&
					buffer[11] == 'V' &&
					buffer[12] == 'E') 
				{
					XHermiteCurve* pc = new XHermiteCurve;
					assert (pc);
					ifs >> *pc;
					m_Objects.push_back(pc);
				}
				else
				// check for bezier patch flag
				// 
				if (buffer[8]  == 'P' &&	//---------------------------- Hermite Patch
					buffer[9]  == 'A' &&
					buffer[10] == 'T' &&
					buffer[11] == 'C' &&
					buffer[12] == 'H') 
				{
					XHermitePatch* pc = new XHermitePatch;
					assert (pc);
					ifs >> *pc;
					m_Objects.push_back(pc);
				}
			}
			else
			{
#ifdef DEBUG
				std::cerr << "[ERROR] FileIO Error while trying to read " 
						  << filename << std::endl;
#endif
				ifs.close();
				return false;
			}
		}
		m_BoneTransforms.resize( m_Bones.size() );
		ifs.close();
	}
	return true;
}
