///-----------------------------------------------------------------------------
/// 
/// \file	Model.h
/// \author	Rob Bateman
/// \brief	This file defines the XModel class. basicaly this class will load the 
///			contents of a mod file. This version merely holds lights and 
///			meshes in one place and is responsible for drawing them.
/// \note	
/// 		This sourcefile is provided for free use for non-commercial purposes only. The
/// 		Software is provided 'as is' and no responsibility can be taken for any damages
/// 		that may be caused while using it. If you make use of this software within 
/// 		your code then it would be nice to have a small credit somewhere in your
/// 		finished product for my hard work :(
/// 		An e-mail saying you found it useful and possibly even a link to your product
/// 		would be really nice too ;)
///
///			Slip me a couple of quid and i might be kind enough to offer assistance in
/// 		sorting out your exporters ;) 
///
///				(C) 2004 Rob 'give me a job offer' Bateman 
/// 
///-----------------------------------------------------------------------------

//////////////////////////////////////////////////////////////////////////
/// The XModel class basically holds the contents of a read file. This may
/// contain a number of 


#ifdef WIN32
#pragma once
#endif

#ifndef __MODEL__H__
#define __MODEL__H__


//////////////////////////////////////////////////////////////////////////

#include <vector>
#include <list>
#include "./AnimCycle.h"
#include "./Bone.h"
#include "./BoneAnimation.h"
#include "./Light.h"
#include "./Material.h"
#include "./Object.h"
#include "./Bone.h"
#include "./TextureFile.h"
#include "./UpdateHook.h"
#include "./SkinCluster.h"
#include "./NurbsCurve.h"
#include "./NurbsSurface.h"

//////////////////////////////////////////////////////////////////////////



/// a list of lights type
typedef std::vector< XBone* > XBoneList;

/// an iterator for a light list
typedef XBoneList::iterator XItBone;


/// a list of lights type
typedef std::vector< XLight* > XLightList;

/// an iterator for a light list
typedef XLightList::iterator XItLight;



/// a list of lights type
typedef std::vector< XObject* > XObjectList;

/// an iterator for a light list
typedef XObjectList::iterator XIt;


/// a list of lights type
typedef std::vector< XTransform > XTransformList;

/// an iterator for a light list
typedef XTransformList::iterator XItTransform;



/// a list of lights type
typedef std::vector< XAnimCycle* > XAnimList;

/// an iterator for a light list
typedef XAnimList::iterator XItAnim;





//////////////////////////////////////////////////////////////////////////

class XModel : protected XUpdateHook
{
	/// grant the deformer access so it can query the bones and objects on this model
	friend class XDeformer;
	friend class XSkinCluster;

public:

	/// ctor
	XModel();

	/// dtor
	~XModel(void);

	/// a function to draw the geometry data in the model
	void gl();

	/// a function to load a file into the model
	bool Load(const char* const filename);

	/// is the model textured?
	bool IsTextured() const;

	/// does the model contain lights?
	bool IsLit() const;
	
	/// \brief	The update hook virtual func
	/// \param	dt	-	the time increment since last frame
	/// 
	void Update(const float dt);

	/// \brief	returns the number of animations that the model has loaded
	/// \return	the number of anim cycles
	/// 
	unsigned int NumAnimations() const;

	/// \brief	This function sets the current animation to the specified id
	/// \param	id	-	the id number of the animation to set as the current anim
	/// \note	this is temporary until I get round to replacing it with blending 
	/// 
	void SetCurrentAnimation(unsigned int id);

	/// \brief	Parents the specified child index to the specified parent transform
	/// \param	child	-	the child shape ID
	/// \param	parent	-	the parent transform ID
	/// \return	true if OK
	/// 
	bool Parent(unsigned int child,unsigned int parent);

private:

	/// \brief	Removes the specified child index from it's parent transform
	/// \param	child	-	the child to remove
	/// \return	true if OK
	/// 
	bool UnParent(unsigned int child);


	/// \brief	This function allows deformers to get access to the bones used 
	///			in the model. 
	/// \param	id	-	the id of the bone to return
	/// \return a pointer to the bone
	/// 
	XBone* GetBone( unsigned int id );

	/// \brief	This function allows deformers to get access to the objects used 
	///			in the model. 
	/// \param	id	-	the id of the object to return
	/// \return a pointer to the object
	/// 
	XObject* GetObject( unsigned int id );
	
protected:

	/// this Function enables all of the lights in the model
	void EnableLights();

	/// this function disables all of the lights in the model
	void DisableLights();

	/// a texture object for all of the textures in the model
	unsigned int m_TextureObject;

	/// the number of textures in the model
	unsigned int m_NumTextures;

	/// a list of materials
	XMaterialList m_Materials;

	/// a list of lights
	XLightList m_Lights;

	/// a list of shapes/cameras/lights/surfaces
	XObjectList m_Objects;

	/// an array of the models joints
	XBoneList m_Bones;
	

	/// an array of bonbe transformations 
	XTransformList m_BoneTransforms;

	/// an array of animation cycles
	XAnimList m_Anims;


	ChildArray m_ParentedObjects;
	
	
	/// a pointer to the current animation cycle.
	/// this is only for testing purposes and will be replaced shortly
	/// with an animation mixer that will do the anim blending....
	/// 
	XAnimCycle* pCurrentAnimation;
};


#endif


