///-----------------------------------------------------------------------------
/// 
/// \file	NurbsCurve.cpp
/// \author	Rob Bateman
/// \brief	This file defines the class XNurbsCurve whixh is used to hold information
///			about a NurbsCurve. 
/// \note	
/// 		This sourcefile is provided for free use for non-commercial purposes only. The
/// 		Software is provided 'as is' and no responsibility can be taken for any damages
/// 		that may be caused while using it. If you make use of this software within 
/// 		your code then it would be nice to have a small credit somewhere in your
/// 		finished product for my hard work :(
/// 		An e-mail saying you found it useful and possibly even a link to your product
/// 		would be really nice too ;)
///
///			Slip me a couple of quid and i might be kind enough to offer assistance in
/// 		sorting out your exporters ;) 
///
///				(C) 2004 Rob 'give me a job offer' Bateman 
/// 
///-----------------------------------------------------------------------------

#include "NurbsCurve.h"
#ifdef WIN32
	#define WIN32_LEAN_AND_MEAN
	#include <Windows.h>
#endif
#include <GL/gl.h>


//-----------------------------------------------------------------------------
/// \brief	ctor
///
XNurbsCurve::XNurbsCurve() 
: XParametricCurve(),m_Degree(0) 
{
}

//-----------------------------------------------------------------------------
/// \brief	dtor
/// 
XNurbsCurve::~XNurbsCurve() 
{
}


//-----------------------------------------------------------------------------
/// \brief	Evaluates the position on the curve at parametric value T. This can 
///			range between the maximum and minimum knot values
/// \perem	T	-	the parametric value
/// \return	The point on the curve at that point
///
void XNurbsCurve::Evaluate(XVertList* UseVerts)
{

	float *pout = m_VertexArray;
	float *pend = pout + GetLOD() * 3;
	std::vector<float>::iterator itb = m_BlendFuncs.begin();

	for ( ; pout != pend; )
	{
		XVertIt it  = UseVerts->begin();
		XVertIt end = UseVerts->end();
		XVector V(0,0,0);

		for (; it != end; ++it, ++itb )
		{
			V += (*itb) * (*it);
		}
		*pout = V.x;
		++pout;
		*pout = V.y;
		++pout;
		*pout = V.z;
		++pout;
	}
}

//-----------------------------------------------------------------------------
/// \brief	This function is called by the XParametricBase::SetLod() function
/// 
void XNurbsCurve::OnSetLod() 
{
	m_BlendFuncs.resize( GetLOD() * m_CVs.size() );

	// delete old vertices
	delete [] m_VertexArray;

	// allocate vertex data
	m_VertexArray = new float[3*GetLOD()];

	std::vector<float>::iterator it = m_BlendFuncs.begin();


#define __EP 0.00001f

	// pre-cache blending factors for U surface direction
	float u_start = m_Knots[m_Degree];
	float u_end   = m_Knots[m_Knots.size()-1-m_Degree];
	float u_total = u_end - u_start;
	float u_inc   = u_total/(GetLOD()-1);
	float u=u_start+__EP;

	for (int i=0;i!=GetLOD();++i,u+=u_inc)
	{
		if(u>=(u_end-__EP)) 
			u = u_end-__EP;

		for (int j=0;j!=m_CVs.size();++j,++it)
		{
			*it = CoxDeBoor(u,j,m_Degree+1,m_Knots);
		}
	}

}


//-----------------------------------------------------------------------------
/// \brief	Reads the Nurbs Curve data
/// \param	ifs	-	the input file stream
/// \param	nc	-	the curve to write to
/// \return	the file stream
/// 
std::istream &operator>>(std::istream& ifs,XNurbsCurve& nc) 
{
	char buffer[64];
	ifs >> buffer; // name
	ifs >> buffer; // "id"
	ifs >> nc.m_ID;
	ifs >> buffer; // "id"
	ifs >> nc.m_ParentID;
	ifs >> buffer; 
	ifs >> nc.m_Degree; 
	ifs >> buffer; 
	
	int NumKnots;
	ifs >> NumKnots;
	
	while(NumKnots--)
	{
		float f;
		ifs >> f;
		nc.m_Knots.push_back(f);
	}

	int NumCvs=0;
	ifs >> buffer >> NumCvs;
	nc.m_CVs.resize(NumCvs);
	for( int i=0;i!=NumCvs;++i)
	{
		ifs >> nc.m_CVs[i];
	}

	nc.SetLOD(20);

	return ifs;
}



