///-----------------------------------------------------------------------------
/// 
/// \file	NurbsSurface.cpp
/// \author	Rob Bateman
/// \brief	This file defines the class XNurbsSurface whixh is used to hold information
///			about a NurbsSurface. 
/// \note	
/// 		This sourcefile is provided for free use for non-commercial purposes only. The
/// 		Software is provided 'as is' and no responsibility can be taken for any damages
/// 		that may be caused while using it. If you make use of this software within 
/// 		your code then it would be nice to have a small credit somewhere in your
/// 		finished product for my hard work :(
/// 		An e-mail saying you found it useful and possibly even a link to your product
/// 		would be really nice too ;)
///
///			Slip me a couple of quid and i might be kind enough to offer assistance in
/// 		sorting out your exporters ;) 
///
///				(C) 2004 Rob 'give me a job offer' Bateman 
/// 
///-----------------------------------------------------------------------------

#include "NurbsSurface.h"
#ifdef WIN32
	#define WIN32_LEAN_AND_MEAN
	#include <Windows.h>
#endif
#include <GL/gl.h>
#include <Vector.h>
#include <Plane.h>
#include <Matrix.h>

XNurbsSurface::XNurbsSurface() 
		: XParametricSurface(),
		  m_DegreeU(0),
		  m_DegreeV(0)
{
}



//-------------------------------------------------------------------
/// \brief	evaluates the surface using the provided surface points
/// \param	UseVerts	-	the vertices to use.
/// 
void XNurbsSurface::Evaluate(XVertList* UseVerts) 
{
	// update any edges for the periodic surfaces just incase the
	// skinning has ignored the points.
	CheckForPeriodSurfaces(UseVerts);

	// tesselate surface in one direction to create a set of temporary
	// curves that can be used for the second direction
	GenerateTemporaryVertices(UseVerts);

	// generate the final set of vertices
	GenerateVertices();

	// now generate the normals (i do this innacurately, though its
	// much faster - performance vs quality again....)
	GenerateNormals();
}

//-------------------------------------------------------------------
/// \brief	This is overidden to update the level of detail on the 
///			parametric shape. This will pre-calculate anything that
///			only changes when the LOD changes, this includes :
/// 
///			texture coords
///			vertex array indices
///			the blending factors for U and V
/// 
void XNurbsSurface::OnSetLod()
{
	// num blend funcs = m_NumCvsU * GetLOD()
	m_BlendFuncsU.resize(m_NumCvsU * GetLOD());
	m_BlendFuncsV.resize(m_NumCvsV * GetLOD());

	// resize the temp points array
	m_TempPoints.resize(GetLOD()*m_NumCvsV);

	// delete old vertex array data
	delete [] m_VertexArray;
	delete [] m_NormalArray;
	delete [] m_UvArray;
	delete [] m_IndexArray;

	// resize vertex array data
	m_VertexArray = new float[ 3 * GetLOD() * GetLOD() ];
	m_NormalArray = new float[ 3 * GetLOD() * GetLOD() ];
	m_UvArray     = new float[ 2 * GetLOD() * GetLOD() ];
	m_IndexArray  = new unsigned int[ 2 * GetLOD() * (GetLOD()-1) ];

	// generate data that does not change per frame
	GenerateTexCoords();
	GenerateIndices();

	std::vector<float>::iterator it;

	#define __EP 0.00001f

	// pre-cache blending factors for U surface direction
	float u_start = m_KnotsU[m_DegreeU];
	float u_end   = m_KnotsU[m_KnotsU.size()-1-m_DegreeU];
	float u_total = u_end - u_start;
	float u_inc   = u_total/(GetLOD()-1);
	float u=u_start+__EP;

	it = m_BlendFuncsU.begin();

	for (int i=0;i!=GetLOD();++i,u+=u_inc)
	{
		if(u>=(u_end-__EP)) 
			u = u_end-__EP;

		for (int j=0;j!=m_NumCvsU;++j,++it)
		{
			*it = CoxDeBoor(u,j,m_DegreeU+1,m_KnotsU);
		}
	}

	// pre-cache blending factors for V surface direction
	float v_start = m_KnotsV[m_DegreeV];
	float v_end   = m_KnotsV[m_KnotsV.size()-1-m_DegreeV];
	float v_total = v_end - v_start;
	float v_inc   = v_total/(GetLOD()-1);
	float v=v_start+__EP;

	it = m_BlendFuncsV.begin();

	for (int i=0;i!=GetLOD();++i,v+=v_inc)
	{
		if(v>=(v_end-__EP) )
			v = v_end-__EP;

		for (int j=0;j!=m_NumCvsV;++j,++it)
		{
			*it = CoxDeBoor(v,j,m_DegreeV+1,m_KnotsV);
		}
	} 
}


//-------------------------------------------------------------------
///	\brief  
///	\param   	
///	\return		
///
void XNurbsSurface::CheckForPeriodSurfaces(XVertList* UseVerts)
{
	// Now then now then, the horror. Maya gives us no weightings for
	// periodic surface directions where points overlap. To get around
	// this we can simply copy the post-skinned points into the correct
	// locations.

	// If both surface directions are periodic, then we basically have a 
	// torus. We need to copy strips from both the U and V sides to cover
	// both edges that will not be skinned. We then also have to copy a 
	// small patch of points
	//
	if (m_TypeU == 2 && m_TypeV == 2)
	{
		unsigned int CopyLenU        = m_DegreeU;	//	length of chunk to copy in U cv's
		unsigned int CopyLenV        = m_DegreeV;	//	length of chunk to copy in V cv's

		unsigned int NextLine        = m_NumCvsV;	//	offset to the next line

		unsigned int CopyStartU = m_NumCvsU - m_DegreeU;
		unsigned int CopyStartV = m_NumCvsV - m_DegreeV;

		// first get the nasty square of points out of the way. Copy the first
		// DegreeU * DegreeV control points to the same place at the end of the
		// surface. 
		//
		{
#if 1
			//	offset from the start of the cv's to those that need to be copied
			unsigned int CopyStartOffset = CopyStartU*m_NumCvsV + CopyStartV;
			
			XVertList::iterator it   = UseVerts->begin();
			XVertList::iterator end  = it + CopyLenV;
			XVertList::iterator copy = it + CopyStartOffset;
			unsigned int inc = m_NumCvsV;

			for (;it != end;++it,++copy)
			{
				XVertList::iterator it_temp   = it;
				XVertList::iterator end_temp  = it + CopyLenU*NextLine;
				XVertList::iterator copy_temp = copy;

				for ( ; it_temp != end_temp; it_temp += NextLine,copy_temp += NextLine)
				{
					*copy_temp = *it_temp;
				}
			}
#endif
		}

		// Next we need to copy the remaining control points on the U side
		//
		{
#if 1
			XVertList::iterator it   = UseVerts->begin();
			XVertList::iterator end  = it + CopyLenU*NextLine;
			XVertList::iterator copy = it + CopyStartU*m_NumCvsV;

			for ( ;it != end; it+=NextLine, copy+=NextLine )
			{
				XVertList::iterator it_temp   = it;
				XVertList::iterator end_temp  = it_temp + CopyStartV;
				XVertList::iterator copy_temp = copy;
				for (; it_temp!= end_temp; ++it_temp, ++copy_temp)
				{
					*copy_temp = *it_temp;
				}
			}
#endif
		}


		// Next we need to copy the remaining control points on the V side
		//
		{
#if 1
			XVertList::iterator it   = UseVerts->begin();
			XVertList::iterator end  = it + m_DegreeV;
			XVertList::iterator copy = it + m_NumCvsV - m_DegreeV;
			unsigned int inc = m_NumCvsV;

			for (;it != end;++it,++copy)
			{
				XVertList::iterator it_temp = it;
				XVertList::iterator end_temp  = it + (m_NumCvsU-CopyLenU) * NextLine;
				XVertList::iterator copy_temp = copy;

				for( ; it_temp!=end_temp; copy_temp += NextLine, it_temp += NextLine )
				{
					*copy_temp = *it_temp;
				}
			}
#endif
		}
	}
	else
	if (m_TypeU == 2)
	{
		XVertList::iterator it   = UseVerts->begin();
		XVertList::iterator end  = it + m_DegreeU*m_NumCvsV;
		XVertList::iterator copy = it + (m_NumCvsU - m_DegreeU)*m_NumCvsV;
		unsigned int inc = m_NumCvsV;

		for (;it != end;++it,++copy)
		{
			*copy = *it;
		}
	}
	else
	if (m_TypeV == 2)
	{
		XVertList::iterator it   = UseVerts->begin();
		XVertList::iterator end  = it + m_DegreeV;
		XVertList::iterator copy = it + m_NumCvsV - m_DegreeV;
		unsigned int inc = m_NumCvsV;

		for (;it != end;++it,++copy)
		{
			XVertList::iterator it_temp = it;
			XVertList::iterator copy_temp = copy;

			for(int i=0;i!=m_NumCvsU;++i,copy_temp+=inc,it_temp+=inc)
			{
				*copy_temp = *it_temp;
			}
		}
		
	}
}


//-------------------------------------------------------------------
///	\brief  This function evaluates the vertex positions for the first
///			u-direction tesselation phase. This will output a series of 
///			points which are then used as the control points for the 
///			final tesselation in the V direction.
/// \param	UseVerts	-	a set of possibly deformed CV's to use to 
///							define the surface.
///	
void XNurbsSurface::GenerateTemporaryVertices(XVertList* UseVerts)
{
	// get an iterator to the output data...
	std::vector<XVector>::iterator it = m_TempPoints.begin();
	
	// get an iterator to the input points
	XVertList::const_iterator itv = UseVerts->begin();

	// loop through each curve in the U direction
	for( int i=0;i!=m_NumCvsV; ++i,++itv )
	{
		// get an iterator to the blending funcs in U direction
		std::vector<float>::iterator itb = m_BlendFuncsU.begin();
		
		// generate each output point on curve 
		for( int j=0; j!=GetLOD(); ++j, ++it )
		{
			// temp reference to the control points
			XVertList::const_iterator itt = itv;

			// start with zero point..
			it->Set(0,0,0);

			for( int k=0; k != m_NumCvsU; ++k, itt += m_NumCvsV, ++itb )
			{
				//	*itt	=	the control point
				//	*itb	=	the blending factor for this point
				//	*it		=	output temporary vertex
				//
				// comp quicker than vector mult, going to be more blending
				// funcs of zero than anything else, so may as well test for
				// the blending factor having no effect before calculating 
				// anything
				// 
				if( (*itb) > 0 )
					*it +=  (*itb) * XVector(itt->x,itt->y,itt->z);
			}
		}
	}
}

//-------------------------------------------------------------------
///	\brief  This function evaluates the vertex positions
///	
void XNurbsSurface::GenerateVertices()
{
	// wipe all vertex values to zero...
	memset(m_VertexArray,0,sizeof(float)*3*GetLOD()*GetLOD());

	// pointer used to step through output vertex array
	float* pOut = m_VertexArray;

	// used to keep track of the start point of the current curve
	// we are tesselating.
	std::vector<XVector>::iterator pPoint = m_TempPoints.begin();

	// surface is LOD * LOD output vertices...	therefore expect
	// a for loop....
	// 
	for( int i=0; i != GetLOD(); ++i,++pPoint )
	{ 
		// get an iterator to the blending funcs in U direction
		std::vector<float>::iterator itb = m_BlendFuncsV.begin();

		// for each curve generate the output point 
		for( int j=0; j!=GetLOD(); ++j )
		{
			// get an iterator for the start of the line
			std::vector<XVector>::iterator pvert = pPoint;

			// loop through each vertex in this curve
			for( int k=0;k!=m_NumCvsV; ++k, ++itb, pvert += GetLOD() )
			{
				// blending factor for this point
				float B = *itb;

				// comp quicker than vector mult, going to be more blending
				// funcs of zero than anything else, so may as well test for
				// the blending factor having no effect before calculating 
				// anything
				// 
				if( B > 0 ) 
				{
					// get quick pointer to output data
					float *p = pOut;

					// add the weighted sum of the points
					*p += B * pvert->x;		++p;
					*p += B * pvert->y;		++p;
					*p += B * pvert->z;
				}
			}

			// move on a bit in the output array
			pOut += 3;
		}		
	}
}


std::istream &operator>>(std::istream& ifs,XNurbsSurface& nc) 
{
	char buffer[64];
	ifs >> buffer; // name
	ifs >> buffer; // "id"
	ifs >> nc.m_ID;
	ifs >> buffer; // "id"
	ifs >> nc.m_ParentID;
	ifs >> buffer; // "material"
	ifs >> nc.m_Material;
	ifs >> buffer; 
	int iii;
	ifs >> iii;
	nc.m_DegreeU=iii; 
	ifs >> buffer; 
	ifs >> iii;
	nc.m_DegreeV=iii; 
	ifs >> buffer; 
	ifs >> iii;
	nc.m_TypeU = iii; 
	ifs >> buffer; 
	ifs >> iii;
	nc.m_TypeV = iii; 
	
	int NumKnots;
	ifs >> buffer; 
	ifs >> NumKnots;
	while(NumKnots--)
	{
		float f;
		ifs >> f;
		nc.m_KnotsU.push_back(f);
	}
	
	ifs >> buffer; 
	ifs >> NumKnots;
	while(NumKnots--)
	{
		float f;
		ifs >> f;
		nc.m_KnotsV.push_back(f);
	}

	ifs >> buffer; 
	ifs >> nc.m_NumCvsU;
	ifs >> nc.m_NumCvsV;
	
	nc.m_CVs.resize(nc.m_NumCvsU*nc.m_NumCvsV);
	for( int i=0;i!=(nc.m_NumCvsU*nc.m_NumCvsV);++i)
	{
		ifs >> nc.m_CVs[i];
	}

	nc.SetLOD(35);

	return ifs;
}
