///-----------------------------------------------------------------------------
/// 
/// \file	NurbsSurface.h
/// \author	Rob Bateman
/// \brief	defines the structures to hold a NurbsSurface. The surface is evaluated
/// 		using the Cox-De-Boor algorithm.
/// \note	
/// 		This sourcefile is provided for free use for non-commercial purposes only. The
/// 		Software is provided 'as is' and no responsibility can be taken for any damages
/// 		that may be caused while using it. If you make use of this software within 
/// 		your code then it would be nice to have a small credit somewhere in your
/// 		finished product for my hard work :(
/// 		An e-mail saying you found it useful and possibly even a link to your product
/// 		would be really nice too ;)
///
///			Slip me a couple of quid and i might be kind enough to offer assistance in
/// 		sorting out your exporters ;) 
///
///
///			(C) 2004 Rob 'I would like a job in Australia or US' Bateman 
/// 
///-----------------------------------------------------------------------------


//------------------------------------------------------------------------------

#ifdef WIN32
	#pragma once
#endif


#ifndef XNURBS_SURFACE__H_
#define XNURBS_SURFACE__H_

	//--------------------------------------------------------------------------

	#ifdef WIN32
		#define WIN32_LEAN_AND_MEAN
		#include <Windows.h>
	#endif

	#include <GL/gl.h>

	#include <vector>
	#include <fstream>
	#include <Vector.h>
	#include <Transform.h>
	#include <Plane.h>

	#include "Material.h"
	#include "Object.h"
	#include "ParametricSurface.h"

	//-----------------------------------------------------------------------	XNurbsSurface
	/// \brief	Defines a class to load and render a nurbs surface. 
	/// 
	class XNurbsSurface 
		: public XParametricSurface
	{

		friend std::istream &operator>>(std::istream&,XNurbsSurface&);

	public:

		//-------------------------------------------------------------------
		/// \brief	ctor
		///
		XNurbsSurface();

		//-------------------------------------------------------------------
		/// \brief	dtor
		///
		~XNurbsSurface() {}

		//-------------------------------------------------------------------
		/// \brief	gets the type of the object
		/// \return	the type of the surface.
		///
		virtual XFn::Type GetType() const {
			return XFn::NurbsSurface;
		}
		
	private:

		//------------------------------------------------------------------- Parametric Interface functions

		//-------------------------------------------------------------------
		/// \brief	This is overidden to update the level of detail on the 
		///			parametric shape
		/// 
		virtual void OnSetLod();

		//-------------------------------------------------------------------
		///	\brief  Evaluates the current surface positions and normals
		///
		virtual void Evaluate(XVertList*);


		//------------------------------------------------------------------- Hard working functions !

		//-------------------------------------------------------------------
		///	\brief  This function evaluates the vertex positions
		///	
		void GenerateVertices();


		//-------------------------------------------------------------------
		///	\brief  This function evaluates the vertex positions for the 
		///			intermediate surface curves.
		///	\param	UseVerts	-	an array of vertices to use for the 
		///							control vertices.
		/// 
		void GenerateTemporaryVertices(XVertList*);

		//-------------------------------------------------------------------
		///	\brief  One little problem with Maya is that when you get a closed
		///			or periodic surface, maya clamps overlapping CV's together.
		///			This means that maya ignores skinning on these CV's so you
		///			end up with points evaluating at zero. To counter this, 
		///			if we know a surface is periodic in any direction, we can
		///			copy the points over prior to being rendered.
		///	\param  UseVerts	-	the vertex coords to use
		///
		void CheckForPeriodSurfaces(XVertList* UseVerts);

		//------------------------------------------------------------------- Tesselation data
		//	To improve the rendering speed of the 
		//	nurbs surfaces, i am pre-caching most 
		//	of the intermediate data. This will 
		//	help to eradicate the costly Cox-De-Boor
		//	evaluation for the surface when drawing
		//	the surface each frame. This has the 
		//  downside in that all of that calculation 
		//	must happen when the Level of detail 
		//	changes....
		//

		/// num blend funcs = m_NumCvsU * GetLOD()
		/// 
		/// This stores the result of the Cox-De-Boor evaluation
		/// for this Surface LOD
		std::vector<float> m_BlendFuncsU;

		/// num blend funcs = m_NumCvsV * GetLOD()
		/// 
		/// This stores the result of the Cox-De-Boor evaluation
		/// for this Surface LOD
		std::vector<float> m_BlendFuncsV;

		/// an array of temporary vertices for the V calculation.
		/// The first stage is to interpolate across the U direction,
		/// the results of which are then stored in the following
		/// array for the surface evaluation in the other direction.
		/// 
		///	In essence this will hold a set of curves across the 
		/// U direction.
		/// 
		std::vector<XVector> m_TempPoints;

		//------------------------------------------------------------------- Surface Data

		/// the knot vector in U direction
		std::vector<float> m_KnotsU;
		
		/// the knot vector in U direction
		std::vector<float> m_KnotsV;

		/// the curve degree in the U direction
		unsigned char m_DegreeU;
		
		/// the curve degree in the V direction
		unsigned char m_DegreeV;

	};


#endif

