//-----------------------------------------------------------------------------
/// 
/// \file	Object.h
/// \author	Rob Bateman
/// \brief	Defines a crappy base class for surface shape types
/// \note	
/// 		This sourcefile is provided for free use for non-commercial purposes only. The
/// 		Software is provided 'as is' and no responsibility can be taken for any damages
/// 		that may be caused while using it. If you make use of this software within 
/// 		your code then it would be nice to have a small credit somewhere in your
/// 		finished product for my hard work :(
/// 		An e-mail saying you found it useful and possibly even a link to your product
/// 		would be really nice too ;)
///
///			Slip me a couple of quid and i might be kind enough to offer assistance in
/// 		sorting out your exporters ;) 
///
///				(C) 2004 Rob 'give me a job offer' Bateman 
/// 
//-----------------------------------------------------------------------------

#ifndef __OBJECT__H__
#define __OBJECT__H__

#ifdef WIN32

	#pragma once

	#define WIN32_LEAN_AND_MEAN
	#include <windows.h>

#endif

#include <vector>
#include <string>
#include <iostream>
#include <Vector.h>
#include <Matrix.h>
#include <Transform.h>
#include <stdlib.h>
#include <GL/gl.h>
#include <GL/glu.h>

#include "Options.h"

// damn that <windows.h>
#ifdef GetObject
	#undef GetObject
#endif

class XMaterial;

/// a list of material types
typedef std::vector< XMaterial* > XMaterialList;


//-----------------------------------------------------------------------------
/// A class to hold a vertex
/// 
class XVertex : public XVector
{
public:

	//-------------------------------------------------------------------
	/// \brief	ctor
	///
	XVertex() : XVector() {}
	XVertex(XReal a,XReal b,XReal c) : XVector(a,b,c) {}

	//-------------------------------------------------------------------
	/// \brief	dtor
	///
	~XVertex(){}

	//-------------------------------------------------------------------
	/// \brief	renders the vertex
	/// 
	inline void gl() const {
		glVertex3f( x,y,z );
	}
};

// typedefs for readabilities sake
typedef std::vector<XVertex> XVertList;
typedef XVertList::iterator XVertIt;


//-----------------------------------------------------------------------------
/// a class to hold a normal
/// 
class XNormal : public XVector
{
public:

	//-------------------------------------------------------------------
	/// \brief	ctor
	///
	XNormal() : XVector() {}

	//-------------------------------------------------------------------
	/// \brief	dtor
	///
	~XNormal(){}

	//-------------------------------------------------------------------
	/// \brief	renders the normal
	/// 
	inline void gl() const {
		glNormal3f( x,y,z );
	}
};


// typedefs for readabilities sake
typedef std::vector<XNormal> XNormList;
typedef XNormList::iterator XNormIt;


//-----------------------------------------------------------------------------
/// a class to hold a uv texture coord
/// 
class XTexCoord
{
	friend std::istream &operator>>(std::istream&,XTexCoord&);

public:

	//-------------------------------------------------------------------
	/// \brief	ctor
	///
	XTexCoord(){}

	//-------------------------------------------------------------------
	/// \brief	dtor
	///
	~XTexCoord(){}

	//-------------------------------------------------------------------
	/// \brief	copy ctor
	///
	XTexCoord(const XTexCoord& v_) 
		: u(v_.u),v(v_.v) {}

	//-------------------------------------------------------------------
	/// \brief	renders the texture coord
	/// 
	inline void gl() const {
		glTexCoord2f( u,v );
	}

	union {
		
		struct {
			/// the tex coord data
			float u;
			float v;
		};
		float data[2];
	
	};

};

// typedefs for readabilities sake
typedef std::vector<XTexCoord> XTexList;
typedef XTexList::iterator XTexIt;


//-----------------------------------------------------------------------------
/// \brief	This namespace holds the enumerations of the various supported Node
///			types. Some of these may as yet be un-supported, they will all work
///			by the end of the series though ;)
/// 
class XFn 
	: protected XOptions {
public:
	enum Type {
		Texture,
		TexturePak,
		
		Material,
		Blinn,
		Lambert,
		Phong,
		PhongE,
		
		Object,

		Deformer,

		CollisionSphere,
		CollisionBox,
		CollisionCone,
		Locator,
		
		Mesh,
		NurbsCurve,
		NurbsSurface,
		BezierCurve,
		BezierPatch,
		HermiteCurve,
		HermitePatch,

		Emitter,

		
		Light,
		AmbientLight,
		PointLight,
		DirectionalLight,
		SpotLight,
		
		Bone,
		
		BlendShape,
		Lattice,
		SkinCluster,
		
		BoneAnimation,
		AnimCycle,
		Last,
		Max=0xFFFF			
	};

	//-------------------------------------------------------------------
	/// gets the type of the object
	virtual XFn::Type GetType() const = 0;
};

//-----------------------------------------------------------------------
/// \brief	Defines the base class of shapes (ie, anything that can be 
/// 		parented under a transform.
///
class XObject 
	: public XFn
{
public:

	//-------------------------------------------------------------------
	/// ctor
	XObject(void);

	//-------------------------------------------------------------------
	/// dtor
	virtual ~XObject(void);

	//-------------------------------------------------------------------
	/// this returns the ID number of the object
	unsigned int GetID() const {
		return m_ID;
	}

	//-------------------------------------------------------------------
	/// this returns the ID number of the object
	unsigned int GetParentID() const {
		return m_ParentID;
	}

	//-------------------------------------------------------------------
	/// gets the type of the object
	virtual XFn::Type GetType() const {
		return XFn::Object;
	}

	//-------------------------------------------------------------------
	/// returns true if the object has normals (skinning)
	virtual bool HasNormals() const {
		return false;
	}

	//-------------------------------------------------------------------
	/// overload this to draw....
	void gl(XMaterialList& mats) {
		this->gl(mats,this->Points(),this->Normals());
	}

	//-------------------------------------------------------------------
	/// this function draws the mesh data
	/// \param	mats	-	an array of loaded materials
	/// \param	UseVerts	-	provides a new set of vertices to use
	/// \param	UseNorms	-	provides a new set of normals to use
	/// \note	by providing the vertex and normal data, we can support 
	/// 		deformations on the surface before it gets rendered. Its
	/// 		important to realise that the original data remains intact
	/// 		held in the derived objects. The arrays passed in here are
	/// 		deformed copies of the originals. 
	/// 		
	///			Deformation is a massive requirement for our simple engine
	/// 		simply because we want to have the ability to support 
	///			lattices/blend shapes/skinning etc. This requires us to deform
	/// 		various copies of the original data.
	///
	virtual void gl(XMaterialList& mats,XVertList* UseVerts,XNormList* UseNorms) {}

	//-------------------------------------------------------------------
	/// returns the vertices for this model
	virtual XVertList* Points() {
		return 0;
	}

	//-------------------------------------------------------------------
	/// returns the normals if used
	virtual XNormList* Normals() {
		return 0;
	}
	
//protected:

	/// the models ID
	unsigned int m_ID;
	
	/// the parent ID
	unsigned int m_ParentID;
};


#include "Material.h"

#endif


