//-------------------------------------------------------------------
///	\file    	ParametricBase.cpp
///	\date   	2/5/2004   12:24
///	\author		Rob Bateman
///	\brief      This file defines a common set of functions and routines
///				used by dynamic curves and surfaces. The aim with anything
///				parametric is to pre-cache as much as possible, and 
///				to only do as much work as is truly needed...
/// \note   
/// 		This sourcefile is provided for free use for non-commercial purposes only. The
/// 		Software is provided 'as is' and no responsibility can be taken for any damages
/// 		that may be caused while using it. If you make use of this software within 
/// 		your code then it would be nice to have a small credit somewhere in your
/// 		finished product for my hard work :(
/// 		An e-mail saying you found it useful and possibly even a link to your product
/// 		would be really nice too ;)
///
///			Slip me a couple of quid and i might be kind enough to offer assistance in
/// 		sorting out your exporters ;) 
///
///				(C) 2004 Rob 'give me a job offer' Bateman     
///
//-------------------------------------------------------------------

#include "ParametricBase.h"



std::vector<XParametricBase*> XParametricBase::m_DynamicObjects;

/// the near tesselation distance
float XParametricBase::m_fNear=1.5f;

/// the far tesselation distance
float XParametricBase::m_fFar=120.0f;

//-------------------------------------------------------------------
/// ctor
///
XParametricBase::XParametricBase(XObject* Obj) 
	: m_pObject(Obj),
	  m_bLocked(false),
	  m_MaxLOD(65),
	  m_bCameraDependant(true)

{
	Insert(this);
}

//-------------------------------------------------------------------
/// dtor
///
XParametricBase::~XParametricBase(void)
{
	Remove(this);
}

//-------------------------------------------------------------------
/// \brief	This function calculates a curve's blending factor for the specified
/// 		point i, given the parametric value u, order k (degree+1) and the given
///		knot vector Knots.
/// \param	u	-	parametric value
/// \param	i	-	the control point idx
/// \param	k	-	the curve order
/// \param	Knots	-	the knot vecctor
/// \return	The blending factor
///
float XParametricBase::CoxDeBoor(float u,int i,int k,const std::vector<float>& Knots)
{
	// if terminating recursion
	if(k==1)
	{
		// if u is in time range, then return 1
		if( Knots[i] <= u && u <= Knots[i+1] ) {
			return 1.0f;
		}

		// otherwise return 0 - this has the effect of removing any
		// influence from point i along the curve because it's too far
		// from the point being evaluated
		return 0.0f;
	}

	// calculate the denominator of the first term
	float Den1 = Knots[i+k-1] - Knots[i];

	// calculate the denominator of the second term
	float Den2 = Knots[i+k] - Knots[i+1];

	// results of the two eqautions
	float Eq1=0,Eq2=0;

	// if denominator of 1st term is 0, then whole term is zero
	if(Den1!=0) {
		Eq1 = ((u-Knots[i]) / Den1) * CoxDeBoor(u,i,k-1,Knots);
	}

	// if denominator of 2nd term is 0, then whole term is zero
	if(Den2!=0) {
		Eq2 = (Knots[i+k]-u) / Den2 * CoxDeBoor(u,i+1,k-1,Knots);
	}

	// result is the sum of the two equations
	return Eq1+Eq2;
}

//-------------------------------------------------------------------
/// \brief	This function must be called every frame to update the
///			position of the camera so that any parametric surfaces
///			tesselated can have their LOD's changed via the camera
///			position. 
/// \param	eye	-	the cameras eye position
/// \param	aim	-	the cameras aim cirection
/// \note	The test performed by this function is somewhat cack.
///			Only the distance between 2 points is tested and does
///			not take into account the aim direction, or surface
///			size/shape. This could do with some form of test such
///			as simple sphere test etc.
/// 
void XParametricBase::SetCameraPosition(const XVector& eye,const XVector& aim)
{
	std::vector<XParametricBase*>::iterator it = m_DynamicObjects.begin();
	for (; it != m_DynamicObjects.end(); ++it)
	{
		// if object LOD is dependant on camera distance
		if( (*it)->m_bCameraDependant )
		{
			// determine distance between object and camera position
			float len = ((*it)->Position() - eye).Length(); 

			// if object is nearer than the near plane, clamp resolution to max
			if (len<m_fNear)
			{
				(*it)->SetLOD((*it)->m_MaxLOD);
			}
			else

			// if very far away, clamp at min resolution
			if (len>m_fFar)
			{
				(*it)->SetLOD(6);
			}
			else
			
			// use an inverse square distance ration to determine a fairly 
			// nice LOD for the given distance. A linear relationship here
			// produces a non-linear visual quality on screen, hence the
			// sqaure....
			//
			{
				// total range for LOD change
				float dista = (m_fFar-m_fNear);

				// how far in to the range is the object
				float distb = (len-m_fNear);

				// prevent division by zero
				if (distb < 0.001f)
				{
					distb=0.1f;
				}

				// ratio value
				float div = 1.0f - distb/dista;

				// calc LOD for object
				float lod = (*it)->m_MaxLOD * ( div * div );

				// set new LOD
				(*it)->SetLOD( lod );
			}
		}
	}
}

//-------------------------------------------------------------------
/// \brief	Sets the near and far tesselation distances forr the
///			camera
///	\param	n	-	the near tesselation distance
///	\param	f	-	the far tesselation distance
/// 
void XParametricBase::SetCameraNearFar(const float& n,const float &f)
{
	m_fNear = n;
	m_fFar  = f;
}

//-------------------------------------------------------------------
///	\brief  Used to insert a pointer into a static parametric surface
///			list.
///	\param  p	-	pointer to the parametric object
///
void XParametricBase::Insert(XParametricBase* p) {
	m_DynamicObjects.push_back(p);
}

//-------------------------------------------------------------------
///	\brief  Used to remove a pointer to a parametric surface
///	\param  p	-	pointer to the parametric object
///
void XParametricBase::Remove(XParametricBase* p) {
	std::vector<XParametricBase*>::iterator it = m_DynamicObjects.begin();
	for (; it != m_DynamicObjects.end(); ++it)
	{
		if (p == *it)
		{
			m_DynamicObjects.erase(it);
			return;
		}
	}
}

//-------------------------------------------------------------------
/// \brief	
void XParametricBase::SetObjectPosition(XObject* Obj,const XVector& V)
{
	std::vector<XParametricBase*>::iterator it = m_DynamicObjects.begin();
	for (; it != m_DynamicObjects.end(); ++it)
	{
		if ( Obj == (*it)->m_pObject )
		{
			(*it)->m_Position = V;
		}
	}
}


