//-------------------------------------------------------------------
///	\file    	ParametricBase.h
///	\date   	1/5/2004   11:56
///	\author		Rob Bateman
///	\brief      
/// \note       
///
//-------------------------------------------------------------------

#ifndef __ParametricBase__H__
#define __ParametricBase__H__

#include <Vector.h>
#include <vector>

#include "Object.h"

class XParametricBase
{
public:

	//---------------------------------------------------------------
	/// ctor
	XParametricBase(XObject* Obj);

	//---------------------------------------------------------------
	/// dtor
	~XParametricBase(void);

	//---------------------------------------------------------------
	/// \brief	This function must be called every frame to update the
	///			position of the camera so that any parametric surfaces
	///			tesselated can have their LOD's changed via the camera
	///			position. 
	/// \param	eye	-	the cameras eye position
	/// \param	aim	-	the cameras aim cirection
	/// 
	static void SetCameraPosition(const XVector&,const XVector&);

	static void SetCameraNearFar(const float& n,const float &f);

	static void SetObjectPosition(XObject* Obj,const XVector& V);

	//---------------------------------------------------------------
	/// returns the level of detail of the shape
	/// 
	const unsigned int& GetLOD() const {
		return m_LOD;
	}

	//---------------------------------------------------------------
	/// \brief	allows you to set the Level of detail for this shape
	/// \param	LOD	-	the new level of detail for the shape
	/// 
	void SetLOD(unsigned int LOD) {
		if (LOD == m_LOD)
		{
			return;
		}
		if(LOD<6) 
			LOD=6;
		m_LOD = LOD;
		OnSetLod();
		if( Locked() ) {
			Evaluate(m_pObject->Points());
		}
	}

	//---------------------------------------------------------------
	/// \brief	locks the surface so that no changes to it's output
	///			data will be made. By default a surface is locked if
	///			it is simply parented under a surface. If any deformer
	///			is applied then it is unlocked to allow for updates.
	///
	void LockSurface() {
		m_bLocked = true;
	}

	//---------------------------------------------------------------
	/// \brief	Unlocks the surface for deformation support
	/// 
	void UnLock() {
		m_bLocked = false;
	}

	//---------------------------------------------------------------
	/// allows you to see if the surface is locked
	/// \return	true if the surface is locked
	const bool Locked() const {
		return m_bLocked;
	}

	const XVector& Position() const {
		return m_Position;
	}
	void SetPosition(const XVector& P) {
		m_Position=P;
	}

protected:

	/// the shapes level of detail
	unsigned int m_LOD;

	//---------------------------------------------------------------
	/// overide this to update the surface
	virtual void Evaluate(XVertList* UseVerts) = 0;

	//---------------------------------------------------------------
	/// overide this function to setup the shapes for a new level of detail
	virtual void OnSetLod() = 0;


	XVector m_Position;

	//---------------------------------------------------------------
	/// \brief	This function calculates a curve's blending factor for the specified
	/// 		point i, given the parametric value u, order k (degree+1) and the given
	///		knot vector Knots.
	/// \param	u	-	parametric value
	/// \param	i	-	the control point idx
	/// \param	k	-	the curve order
	/// \param	Knots	-	the knot vecctor
	/// \return	The blending factor
	///
	static float CoxDeBoor(float u,int i,int k,const std::vector<float>& Knots);

	/// the maximum possible LOD for the surface when it is at the
	/// NEAR distance from the camera
	unsigned int m_MaxLOD;

private:

	static std::vector<XParametricBase*> m_DynamicObjects;

	//-------------------------------------------------------------------
	///	\brief  Used to insert a pointer into a static parametric surface
	///			list.
	///	\param  p	-	pointer to the parametric object
	///
	static void Insert(XParametricBase* p);

	//-------------------------------------------------------------------
	///	\brief  Used to remove a pointer to a parametric surface
	///	\param  p	-	pointer to the parametric object
	///
	static void Remove(XParametricBase* p);

	/// the near tesselation distance
	static float m_fNear;

	/// the far tesselation distance
	static float m_fFar;

	/// a flag to lock the 
	bool m_bLocked;

	/// if true, the surfaces LOD will be dynamically tesselated between 
	/// the near and far viewing distance of the camera. Rah!
	bool m_bCameraDependant;



	/// pointer to itself...
	XObject* m_pObject;
};

#endif

