///-----------------------------------------------------------------------------
/// 
/// \file	SkinCluster.h
/// \author	Rob Bateman
/// \brief	This file defines the XAnimCycle class. basicaly this class holds a 
///			set of animation cycles for each bone in the hierarchy. 
/// 
///-----------------------------------------------------------------------------

//////////////////////////////////////////////////////////////////////////
/// The XModel class basically holds the contents of a read file. This may
/// contain a number of 

#ifdef WIN32
	#pragma once
#endif

#ifndef SKIN_CLUSTER_H_INCLUDED_
#define SKIN_CLUSTER_H_INCLUDED_


// crappy VC6 template warnings
#ifdef WIN32
	#pragma warning(disable:4786)
#endif

#include<vector>
#include<cassert>
#include<string>

#include"Bone.h"
#include"BoneAnimation.h"
#include"Object.h"
#include"Mesh.h"
#include"Model.h"
#include"Deformer.h"


///////////////////////////////////////////////////////////////////////////
/// \brief	The normal reference class exists to specifically associate 
///			a particular normal with a particular weighted vertex. Due to
///			the fun problem of having more than 1 possible normal at any given
///			vertex, the skin weight class holds a number of references to the
///			normals at the vertex. This allows normals & vertices to be
///			transferred for a laugh.
/// 
class XNormalRef
{
public:
	
	///////////////////////////////////////////////////////////////////////
	/// ctor
	XNormalRef( unsigned short ind,
				const XVector& lspaceNormal) 
		: m_NormalID(ind),m_LocalSpaceNormal(lspaceNormal)
	{
	}

	///////////////////////////////////////////////////////////////////////
	/// \brief	This function returns the index of the normal vector
	/// \return	The norml index
	/// 
	inline const unsigned short GetIndex() const {
		return m_NormalID;
	}

	///////////////////////////////////////////////////////////////////////
	/// \brief	This function returns the normal vector in local space
	///			relative to the wieghted transform
	/// \return	The vector
	/// 
	inline const XVector& GetNormal() const {
		return m_LocalSpaceNormal;
	}

protected:

	/// this is the normal in local space relative to the weighted transform
	XVector				m_LocalSpaceNormal;

	/// this is the index to the related normal
	unsigned short		m_NormalID;
};



///////////////////////////////////////////////////////////////////////////
/// \brief	An individual skin weight
/// 
class XSkinWeight
{
	friend class XSkinCluster;

public:

	/// ctor
	XSkinWeight() 
		: m_fWeight(0),
		  m_BoneID(0),
		  m_VertexID(0),
		  m_LocalSpaceVert()
	{
	}
	/// ctor
	XSkinWeight(const XSkinWeight& w) 
		: m_fWeight(w.m_fWeight),
		  m_BoneID(w.m_BoneID),
		  m_VertexID(w.m_VertexID),
		  m_LocalSpaceVert(w.m_LocalSpaceVert)
	{
	}
	
	///////////////////////////////////////////////////////////////////////
	/// \brief	This function returns the bone that relates to this weight
	/// \return	the bone ID
	/// 
	unsigned short GetBoneID() const {
		return m_BoneID;
	}

	///////////////////////////////////////////////////////////////////////
	/// \brief	This simply sets the bone ID for this weight
	/// \param	id	-	the bone ID to use
	/// 
	void SetBoneID(const unsigned short id)	{
		m_BoneID = id;
	}

	///////////////////////////////////////////////////////////////////////
	/// \brief	This simply sets the bone  weight
	/// \param	w	-	the weight to use
	/// 
	void SetWeight(const float w) {
		m_fWeight = w;
	}

	///////////////////////////////////////////////////////////////////////
	/// \brief	This sets the local space vertex coord for this weighted vert
	/// \param	id	-	the vertex ID
	/// \param	v	-	the new local space vertex
	/// 
	void SetVertex(const unsigned short id,const XVector& v) {
		m_VertexID = id;
		m_LocalSpaceVert = v;
	}

//protected:

	/// the skin weight for this bone
	float						m_fWeight;

	/// the bone ID
	unsigned short				m_BoneID;

	/// the vertex id that this weight applies to
	unsigned short				m_VertexID;

	/// the vertex coord in local space
	XVector						m_LocalSpaceVert;

};


///////////////////////////////////////////////////////////////////////////
/// \brief	The skin cluster holds skinning information that applies a 
/// 		
class XSkinCluster 
	: public XDeformer
{
	typedef std::vector<XSkinWeight> WeightList_;
	typedef WeightList_::iterator WeightIt_;

	friend std::istream &operator>>(std::istream&,XSkinCluster&);
public:


	/// gets the type of the object
	virtual XFn::Type GetType() const {
		return XFn::SkinCluster;
	}
	XSkinCluster(XModel* M) 
		: XDeformer(M),
		  m_pTheObject(0){}

	///////////////////////////////////////////////////////////////////////
	/// \brief	This function is designed to apply the skinning info to the 
	///			final mesh
	/// \param	rot_xforms	-	an array of transformed and deformed vertices
	/// \param	xforms		-	an array of transformed normal vectors
	/// \param	OutputVertices	-	an array of output vertices
	/// \param	OutputNormals	-	an array of output normals
	/// 
	void ApplySkinning( XVertList& OutputVertices );


	///////////////////////////////////////////////////////////////////////
	/// \brief	This is an exceptionally minging function, unfortunately
	///			due to the skin cluster having various dependencies with meshes
	///			and bones etc, I cannot use the streaming operators here because
	///			I need to perform a wee bit of calculation first.....
	/// \param	ifs	-	the input file stream
	/// \param	BoneIDs	-	an array for resolution of the correct bone ID
	/// \param	WorldSpaceBoneMatrices	-	an array of all world space matrices of bones
	/// \param	mesh	-	the mesh to which the skin cluster is applied
	/// 
	void ReadWeights( std::istream &ifs,
					  std::vector<unsigned int>& BoneIDs,
					  std::vector<XMatrix>& WorldSpaceBoneMatrices,
					  const XMesh& mesh);

	///////////////////////////////////////////////////////////////////////
	/// this function draws the mesh data
	/// \param	mats	-	an array of loaded materials
	/// \param	UseVerts	-	provides a new set of vertices to use
	/// \param	UseNorms	-	provides a new set of normals to use
	/// \note	by providing the vertex and normal data, we can support 
	/// 		deformations on the surface before it gets rendered. Its
	/// 		important to realise that the original data remains intact
	/// 		held in the derived objects. The arrays passed in here are
	/// 		deformed copies of the originals. 
	/// 		
	///			Deformation is a massive requirement for our simple engine
	/// 		simply because we want to have the ability to support 
	///			lattices/blend shapes/skinning etc. This requires us to deform
	/// 		various copies of the original data.
	///
	virtual void gl( XMaterialList& mats       ,
					 XVertList* UseVerts ,
					 XNormList* UseNorms ) {
		ApplySkinning( m_Points );

		m_pTheObject->gl(mats,&m_Points,m_pTheObject->Normals());

	}

	///////////////////////////////////////////////////////////////////////
	/// \brief	returns a pointer to the point list for this surface
	///
	virtual XVertList* Points() {
		return &m_Points;
	}

	void CalculateBasePose();

//protected:

	void DoCalc(XBone* B,const XMatrix& WS = XMatrix() );

	int GetBoneIdx(XBone* B);

	///////////////////////////////////////////////////////////////////////
	/// \brief	Sets the skinned object
	/// \param	id	-	the id number of the object to set
	/// 
	void SetObject(unsigned int obj) ;

	/// storage for the vertices
	XVertList m_Points;

	/// a pointer to the object that this skin cluster applies to.
	XObject* m_pTheObject;

	std::vector<XMatrix> m_WorldMatrices;
	 	
	/// all of the influences for the skin cluster
	std::vector<unsigned short> m_Influences;
					
	/// an array of skin weights
	WeightList_	m_SkinWeights;
};

typedef std::vector<XSkinCluster*> XSkinList;
typedef XSkinList::iterator XSkinIt;

#endif
