///-----------------------------------------------------------------------------
/// 
/// \file	Texture.h
/// \author	Rob Bateman
/// \brief	This file defines the XImageHeader and the XTexture classes. These 
///			are used to read a texture file from a texture file
/// \note	
/// 		This sourcefile is provided for free use for non-commercial purposes only. The
/// 		Software is provided 'as is' and no responsibility can be taken for any damages
/// 		that may be caused while using it. If you make use of this software within 
/// 		your code then it would be nice to have a small credit somewhere in your
/// 		finished product for my hard work :(
/// 		An e-mail saying you found it useful and possibly even a link to your product
/// 		would be really nice too ;)
///
///			Slip me a couple of quid and i might be kind enough to offer assistance in
/// 		sorting out your exporters ;) 
///
///				(C) 2004 Rob 'give me a job offer' Bateman 
/// 
/// 
///-----------------------------------------------------------------------------

#ifdef WIN32
#pragma once
#endif

#ifndef __X_TEXTURE__H__
#define __X_TEXTURE__H__

#include <stdio.h>
#include <math.h>

#include "Object.h"

///-----------------------------------------------------------------------------
/// \brief	This struct is used as a 2byte identifier just before an image
/// \note	sizeof(ImageHeader) == 2
/// 
struct XImageHeader
{
	/// the width of the image as the power of two
	unsigned char m_WidthPower:4;

	/// the height of the image as a power of two
	unsigned char m_HeightPower:4;

	/// the bits per pixel
	/// 0 = 8bit
	/// 1 = 16bit
	/// 2 = 24bit
	/// 3 = 32bit
	/// 
	unsigned char m_Bpp:2;

	/// 6bits for a rainy day...
	unsigned char m_Unused:6;
};

///-----------------------------------------------------------------------------
/// \brief	a class to hold a texture
/// 
class XTexture
	: public XFn
{
public:

	///-------------------------------------------------------------------------
	/// \brief	ctor
	///
	XTexture();

	///-------------------------------------------------------------------------
	/// \brief	dtor
	///
	~XTexture();

	///-------------------------------------------------------------------------
	/// \brief	returns the type of this scene object, XFn::Texture
	/// \return	the object type
	/// 
	virtual XFn::Type GetType() const {
		return XFn::Texture;
	}

	///-------------------------------------------------------------------------
	/// \brief	returns the width of the texture
	/// \return	the width of the texture
	/// 
	inline unsigned short GetWidth() const {
		return static_cast<unsigned short>(pow(static_cast<unsigned short>(2),m_Header.m_WidthPower));
	}

	///-------------------------------------------------------------------------
	/// \brief	returns the height of the texture
	/// \return	the height of the texture
	/// 
	inline unsigned short GetHeight() const {
		return static_cast<unsigned short>(pow(static_cast<unsigned short>(2),m_Header.m_HeightPower));
	}

	///-------------------------------------------------------------------------
	/// \brief	returns the bits per pixel of the texture
	/// \return	the bits per pixel of the texture
	/// 
	inline unsigned short GetBitsPerPixel() const {
		return GetBytesPerPixel() << 3;
	}

	///-------------------------------------------------------------------------
	/// \brief	returns the bytes per pixel of the texture
	/// \return	the bytes per pixel of the texture
	///
	inline unsigned short GetBytesPerPixel() const {
		return 1 + m_Header.m_Bpp;
	}

	///-------------------------------------------------------------------------
	/// \brief	returns a pointer to the pixel data array
	/// \return	address of the first pixel
	/// 
	inline const unsigned char* const GetPixels() const {
		return m_Pixels;
	}

	///-------------------------------------------------------------------------
	/// \brief	this function reads a texture from the specified file pointer
	/// \param	fp	-	the file to read from
	/// \return	true if OK
	/// 
	bool Read(FILE* fp);

protected:

	/// the image header
	XImageHeader   m_Header;

	/// a pointer to the pixel data
	unsigned char* m_Pixels;
};

#endif


