///-----------------------------------------------------------------------------
/// 
/// \file	UpdateHook.h
/// \author	Rob Bateman
/// \brief	Defines a simple update mechanism
/// \note	
/// 		This sourcefile is provided for free use for non-commercial purposes only. The
/// 		Software is provided 'as is' and no responsibility can be taken for any damages
/// 		that may be caused while using it. If you make use of this software within 
/// 		your code then it would be nice to have a small credit somewhere in your
/// 		finished product for my hard work :(
/// 		An e-mail saying you found it useful and possibly even a link to your product
/// 		would be really nice too ;)
///
///			Slip me a couple of quid and i might be kind enough to offer assistance in
/// 		sorting out your exporters ;) 
///
///				(C) 2004 Rob 'give me a job offer' Bateman 
/// 
///-----------------------------------------------------------------------------

#ifdef WIN32
#pragma once
// crappy VC6 template warnings
#pragma warning(disable:4786)
#endif

#ifndef __X__UPDATE__HOOK__H_
#define __X__UPDATE__HOOK__H_


#include <set>
#include <cassert>
#include <algorithm>

// forward declarations....
class XUpdateHook;


/// the internal hook type is used as nothing more than a wrapper for 
/// an XUpdateHook pointer. This impliments the basic < operator for
/// priority sorting, mainly the reason why I'm using std::set cos I
/// can't be arsed to do it manually. I suppose std::sort and a vector 
/// would also do the trick but probably a bit more optimal, oh well...
/// 
struct XInternalHook 
{


	/// various ctors for use with std::set
	XInternalHook() : m_ptr(NULL) {}
	XInternalHook(XUpdateHook* ptr) : m_ptr(ptr) {}
	XInternalHook(const XInternalHook& rhs) : m_ptr(rhs.m_ptr) {}

public:
	
	/// this function passes te update call from the reference class to the 
	/// update hook class
	/// 
	inline void Update(const float dt);

	// needed for std::set so it can sort based on priority
	inline bool operator < (const XInternalHook& rhs) const;

private:
	
	/// the internal pointer to the update class
	XUpdateHook* m_ptr;
};

class XUpdateHook
{
public:

	/// ctor	-	adds node into internal set
	XUpdateHook(void);

	/// dtor	-	removes element from update queue
	~XUpdateHook(void);
	
	/// \brief	This Enables the update. It's likely that you will call other
	///			function Enable and Disable, if you have, call XUpdateHook::Enable()
	///		
	void Enable();

	/// \brief	This disables the update. It's likely that you will call other
	///			function Enable and Disable, if you have, call XUpdateHook::Disable()
	///		
	void Disable();

	/// \brief	This function increases the priority of the update.
	/// \note	std::set sorts by the less than operator by default. I could of
	///			overloaded that to do greater than comparisons, however I was
	///			being lazy and decided to just use lower values for higher 
	///			priorities. You don't need to know that when using the class 
	///			however
	void IncPriority() {
		assert( m_Priority != 0x00 );
		--m_Priority;
	}

	/// \brief	This function decreases the priority of the update.
	/// \note	std::set sorts by the less than operator by default. I could of
	///			overloaded that to do greater than comparisons, however I was
	///			being lazy and decided to just use lower values for higher 
	///			priorities. You don't need to know that when using the class 
	///			however
	void DecPriority() {
		assert( m_Priority != 0xFF );
		++m_Priority;
	}


	/// \brief	This function decreases the priority of the update.
	/// \note	std::set sorts by the less than operator by default. I could of
	///			overloaded that to do greater than comparisons, however I was
	///			being lazy and decided to just use lower values for higher 
	///			priorities. You don't need to know that when using the class 
	///			however
	unsigned char GetPriority() {
		return m_Priority;
	}
	

	/// \brief	Derived classes should overload this function to recieve updates
	///	\param	dt	-	the time increment over the frame
	/// 
	virtual void Update(const float dt) = 0;

private:




	/// the type of the set used
	typedef std::set<XInternalHook> SetType_;

	/// the iterator type
	typedef SetType_::iterator SetIt_;

	/// the update queue is implimented as a set so that all elements can be 
	/// sorted nicely.
	static SetType_ ms_UpdateQueue;

	/// this function simply adds a hook into the update queue
	/// \param	ptr	-	pointer to the new hook
	/// 
	static void AddHook(XUpdateHook* ptr) 
	{
		assert(ptr && "Update hook was passed in as NULL" );
		ptr->m_QueuePosition = ms_UpdateQueue.insert( ms_UpdateQueue.begin(), XInternalHook(ptr) );
	}

	static void DeleteHook(XUpdateHook* ptr) 
	{
		assert(ptr && "Update hook was passed in as NULL" );
		ms_UpdateQueue.erase( ptr->m_QueuePosition );
	}
	
	/// the priority for the update. By default this will be 127, or 0x0F. 
	unsigned char m_Priority;

	/// the position of the hook in the update queue
	SetIt_ m_QueuePosition;


};

/// this function passes te update call from the reference class to the 
/// update hook class
/// 
void XInternalHook::Update(float dt) {
	m_ptr->Update(dt);
}

bool XInternalHook::operator < (const XInternalHook& rhs) const {
	return m_ptr->GetPriority() < rhs.m_ptr->GetPriority();
}

#endif

