// EulerAngles.h: interface for the XEulerAngles class.
//
//////////////////////////////////////////////////////////////////////

#ifndef EULER_ANGLES_H_INCLUDED
#define EULER_ANGLES_H_INCLUDED

#ifdef WIN32
#pragma once
#endif

#include "MathDefs.h"
#include "Vector.h"
#include "Matrix.h"
#include "Quaternion.h"

/// \brief	This enumeration defines the order of the euler rotations
/// 
enum XRotationOrder 
{
	XROT_XYZ,
	XROT_XZY,
	XROT_YXZ,
	XROT_YZX,
	XROT_ZXY,
	XROT_ZYX
};

/// \brief	This class defines a euler rotation comprising x y and z rotation values.
///			The euler rotations can also occur in any combination of xyz, denoted by
///			the XrotationOrder enumeration
///
class XEulerAngles  
{
public:

	/// \brief	ctor
	/// \param	ro	-	the order in which the rotations occur.
	/// 
	XEulerAngles(XRotationOrder ro = XROT_XYZ)
		: m_xRotation(0), m_yRotation(0), m_zRotation(0), m_RotationOrder(ro) 
	{
	}

	/// \brief	ctor
	/// \param	x	-	the x rotation
	/// \param	y	-	the y rotation 
	/// \param	z	-	the z rotation
	/// \param	ro	-	the order in which the rotations occur.
	/// 
	XEulerAngles(const XReal x,const XReal y,const XReal z,XRotationOrder ro = XROT_XYZ)
		: m_xRotation(x), m_yRotation(y), m_zRotation(z), m_RotationOrder(ro) 
	{
	}

	/// \brief	dtor
	~XEulerAngles() {};

	inline XReal& RotX() {
		return m_xRotation;
	}
	inline const XReal& RotX() const {
		return m_xRotation;
	}
	inline XReal& RotY() {
		return m_yRotation;
	}
	inline const XReal& RotY() const {
		return m_yRotation;
	}
	inline XReal& RotZ() {
		return m_zRotation;
	}
	inline const XReal& RotZ() const {
		return m_zRotation;
	}

	/// \brief	This function allows you to set the x rotation of the euler
	///			rotation. 
	/// \param	x_	-	the new x value
	///
	inline void SetX(const XReal x_) {
		m_xRotation = x_;
	}
	
	/// \brief	This function allows you to set the x rotation of the euler
	///			rotation. 
	/// \param	y_	-	the new y value
	///
	inline void SetY(const XReal y_) {
		m_yRotation = y_;
	}
	
	/// \brief	This function allows you to set the x rotation of the euler
	///			rotation. 
	/// \param	z_	-	the new z value
	///
	inline void SetZ(const XReal z_) {
		m_zRotation = z_;
	}

	/// \brief	this function returns the x value of the euler rotation
	/// \return	the x rotation value
	///
	inline const XReal GetX() const {
		return m_xRotation;
	}

	/// \brief	this function returns the y value of the euler rotation
	/// \return	the y rotation value
	///
	inline const XReal GetY() const {
		return m_yRotation;
	}

	/// \brief	this function returns the z value of the euler rotation
	/// \return	the z rotation value
	///
	inline const XReal GetZ() const {
		return m_zRotation;
	}

	/// \brief	this function sets the order of the euler rotations
	/// \param	the new rotation order
	///
	inline void SetOrder(const XRotationOrder order) {
		m_RotationOrder = order;
	}

	/// \brief	this function returns the order of the euler rotations
	/// \return	the rotation order
	///
	inline const XRotationOrder GetOrder() const {
		return m_RotationOrder;
	}


	/// \brief	this function returns the euler rotation as a matrix
	/// \return	the matrix that represents the euler rotation
	///
	XMatrix AsMatrix() const;

	/// \brief	this function returns the euler rotation in quaternion
	/// \return	the rotation as a quaternion
	///
	//XQuaternion AsQuaternion() const;

protected:

	/// the x rotation of the transform
	XReal m_xRotation;

	/// the y rotation of the transform
	XReal m_yRotation;

	/// the z rotation of the transform
	XReal m_zRotation;

	/// the rotation order of the transform
	XRotationOrder m_RotationOrder;
};

#endif

