
//=========================================================================
///
///	\file	line.h
/// \author	Rob
///	\brief	Defines the class XLine, a class definition for a 3D line
///
//=========================================================================


//=========================================================================
//	Sentry
//=========================================================================

//	This directive is used to stop the header file from being included
//	Twice and is standard for all headers.
#ifdef WIN32
#pragma once
#endif


#ifndef LINE_H_INCLUDED
#define LINE_H_INCLUDED

//=========================================================================
//	Includes
//=========================================================================

#include "Vector.h"
#include "Matrix.h"

class XLine 
{
public:

	/// ctor
	inline XLine() : Point1(), Point2() {}
	
	/// \brief	ctor
	/// \param	p1	-	the first point on the line
	/// \param	p2	-	the second point on the line
	/// 
	inline XLine(const XVector& p1,const XVector& p2) : Point1(p1), Point2(p2) {}
	
	/// \brief	copy ctor
	/// \param	l	-	the line to copy
	/// 
	inline XLine(const XLine& l) : Point1(l.Point1), Point2(l.Point2) {}

	/// \brief	This function returns a vector from point1 to point2
	/// \return	The vector point2-point1
	/// 
	inline XVector Vector() const {
		return Point2-Point1;
	}

	/// \brief	This function returns the normalised direction vector from 
	///			point1 to point2
	/// \return	The normalised direction vector
	/// 
	inline XVector Direction() const {
		return Vector().Normalise();
	}

	/// \brief	This function returns the length of the line as a real number
	/// \return	The lines length
	/// 
	inline XReal Length() const {
		return Vector().Length();
	}

	/// \brief	This function allows you to transform the line via a matrix
	/// \param	m	-	the matrix transformation
	/// \param	l	-	the line to transform
	/// 
	inline friend XLine operator * (const XMatrix &m,const XLine& l) {
		return XLine( m*l.Point1, m*l.Point2 );
	}

	/// \brief	This function returns the closest point on the line to the 
	///			point specified, p
	/// \param	A point to which you wish to know the closest point on the line
	/// \return	The closest point on the line to p
	/// 
	XVector ClosestPoint(const XVector &p) const ;

	/// the first point defining the line
	XVector Point1;
	
	/// the second point defining the line
	XVector Point2;

};

#endif




