//=========================================================================
///
///	\file	Plane.h
/// \author	Rob
///	\brief	Defines the class XPlane, a class definition for a 3D plane of
///			the form
///
///			ax + by + cz + d = 0
///
//=========================================================================


//=========================================================================
//	Sentry
//=========================================================================
#ifdef WIN32
#pragma once
#endif


#ifndef PLANE_H_INCLUDED
#define PLANE_H_INCLUDED

//=========================================================================
//	Includes
//=========================================================================

#include "MathDefs.h"
#include <iostream>
#include "Vector.h"



//=========================================================================
//	Class Definitions
//=========================================================================

class XPlane
{
public :

	//	bit of a hack
	//	=============
	//
	//	using a union you can share a chunk of memory in order to switch
	//	between different data layouts. In C++ you do not need to specify
	//	the name for an enum, union or structure if it's defined within a
	//	class or structure.
	//
	//	The effect of having no name for  union, struct or enum definitions
	//	is that their contents becomes a part of the parent scope. In this
	//	example then, a,b,c,d and data are in the scope XPlane::
	//
	//	Anyhow, this union is used to allow 2 representations of 16bytes of
	//	memory (enough for 4 floats). The first way is as a 4 float array
	//	called data. The second uses a nameless struct to package four floats
	//	together into a single struct (allows the union to work properly).
	//
	//	All of this means the following, when creating a plane :
	//
	//	XPlane p;
	//
	//	you can access the data with the following :
	//
	//	p.a
	//	p.b
	//	p.c
	//	p.d
	//
	//	you could also do :
	//
	//	p.data[0]
	//	p.data[1]
	//	p.data[2]
	//	p.data[3]
	//
	//	therefore, the folowing are one and the same
	//
	//	p.a		===		p.data[0]
	//	p.b		===		p.data[1]
	//	p.c		===		p.data[2]
	//	p.d		===		p.data[3]
	//
	union
	{
		XReal data[4];
		struct
		{
			XReal a;
			XReal b;
			XReal c;
			XReal d;
		};
	};

    ///	\brief	ctor
	///
	inline XPlane() : a(0), b(0), c(0), d(0)
	{
    }

    ///	\brief	ctor
	///	\param	a_	-	x co-efficient, a, of plane equation
	///	\param	b_	-	y co-efficient, b, of plane equation
	///	\param	c_	-	z co-efficient, c, of plane equation
	///	\param	d_	-	d constant of plane equation
	///
	inline XPlane( IN const XReal a_ ,
				   IN const XReal b_ ,
				   IN const XReal c_ ,
				   IN const XReal d_ )
		: a(a_), b(b_), c(c_), d(d_)
	{
    }

    ///	\brief	ctor
	///	\param	v	-	the surface normal of the plane
	///	\param	D	-	d constant of plane equation
	///
	inline XPlane( IN const XVector &v ,
				   IN const float D )
		: a(v.x), b(v.y),c(v.z),d(D)
	{
    }

    ///	\brief	copy ctor
	///	\param	p	-	the plane to copy
	///
	inline XPlane( IN const XPlane &p )
		: a(p.a), b(p.b),c(p.c),d(p.d)
	{
    }

	/// \brief	ctor - constructs plane equation for the three points v1,v2,v3
	/// \param	v1	-	first point that lies on the plane
	/// \param	v2	-	second point that lies on the plane
	/// \param	v3	-	third point that lies on the plane
	///
    inline XPlane( IN const XVector &v1 ,
				   IN const XVector &v2 ,
				   IN const XVector &v3 )
	{
		XVector n =  ( v3 - v2 ) ^ ( v1 - v2 ) ;
		n.Normalise();
		SetNormal(n);
		d = -(n*v1);
    }

    ///	\brief	dtor
	///
	inline ~XPlane()
	{
	}

	/// \brief	assignment operator  =
	/// \param	p	-	the plane to copy from
	/// \return	reference to itself
	///
    inline const XPlane& operator=( IN const XPlane& p )
	{
		a = p.a; b = p.b; c = p.c; d = p.d;
		return *this;
    }

    /// \brief	returns the normal perpendicular to the surface
	/// \return	the plane normal
	/// \note	the normal vector of the plane is merely (a,b,c)
	///
    inline XVector Normal( void ) const
	{
		return XVector( a , b , c );
    }

    /// \brief	sets the plane equation normal
	/// \param	n	-	the new surface normal
	/// \return	a reference to itself
	/// \note
	///
    inline XPlane &SetNormal( IN const XVector &n )
	{
		a = n.x; b = n.y; c = n.z;
		return *this;
    }

    /// \brief	calculates the shortest distance between the point and
	///			the plane. This will be negative if point is behind the plane
	/// \param	point	-	the point to test
	/// \return	the distance from the plane +ve or -ve
	/// \note
	///
    inline XReal Distance( IN const XVector &point ) const
	{
		return ( Normal() * point ) + d;
    }

    /// \brief	a function to find the intersection point between a plane
	///			and a line. The function returns a distance value (k) with
	///			which you can get the actual point using the following :
	///
	///			intersect_point = origin_of_line  +  k * dir_of_line;
	///
	/// \param	origin		-	the start point of the line
	/// \param	direction	-	the unit length direction vector of the line
	/// \return	the distance to the intersect point
	/// \note
	///
    inline XReal LineIntersection( IN const XVector &origin , IN const XVector &direction ) const
	{
		return -((Normal() * origin) + d)/(Normal() * direction);
    }

    /// \brief	a function to find out if the plane equation is valid
	/// \return	true if equation is OK
	/// \note
	///
    inline bool Valid() const
	{
		float l = Normal().Length();
		return ( ( l > 0.0f ) && ( ( 1.0f/l ) > 0.0f ) );
    }

    /// \brief	a function to set the plane equation by passing it three
	///			points that are known to be on the plane.
	/// \param	v1	-	first point on plane
	/// \param	v2	-	second point on plane
	/// \param	v3	-	third point on plane
	/// \return	a reference to itself
	/// \note
	///
    inline XPlane &Set( IN const XVector &v1 ,
						IN const XVector &v2 ,
						IN const XVector &v3 )
	{
		XVector n = ( ( v3 - v2 ) ^ ( v1 - v2 ) ).Normalise();
		SetNormal(n);
		d = -(n*v1);
		return *this;
    }

    /// \brief	a function to
	/// \param
	/// \param
	/// \param
	/// \param
	/// \return
	/// \note
	///
    inline XPlane &Set( IN const XReal a_ ,
						IN const XReal b_ ,
						IN const XReal c_ ,
						IN const XReal d_ )
	{
		a=a_;b=b_;c=c_;d=d_;
		return *this;
    }

	/// This function returns the intersect point between a plane and a ray. The ray
	/// is defined by it's starting point a, and the normalised direction vector dir
	/// \return	The intersection point.
	/// \note	This function does not actually test to see if an intersection is
	///			actually possible. God knows what will happen if the ray never actually 
	///			intersects the plane!
	/// 
	inline XVector GetRayIntersect( const XVector& a, const XVector& dir ) const {
		// distance from plane * direction == line
		return a + dir*Distance(a);	
	}

	/// This function returns the intersect point between a plane and a line. The line
	/// is defined between two points; a & b
	/// \return	The intersection point.
	/// \note	This function does not actually test to see if an intersection is
	///			actually possible. God knows what will happen if the ray never actually 
	///			intersects the plane!
	/// 
	inline XVector GetLineIntersect( const XVector& a, const XVector& b ) const {
		// end point - start point == direction.
		// normalise it
		XVector dir = (b-a).Normalise();

		// distance from plane * direction == line
		return GetRayIntersect(a,dir);	
	}

	/// This function intersects the polygon defined by inpoints by this plane. The 
	/// resulting polygon is stored within outpoints. 
	/// 
	void CutPoly(XVecList &outpoints,const XVecList& inpoints) const;

};

#endif
