
#include "Sphere.h"


	/// \brief	This function is used to get hold of the line intersection points when a line
	///			intersects a sphere. There are 3 possibilities, either
	///				1. the line does not intersect the sphere
	///				2. The line intersects the sphere once (one point is inside the sphere)
	///				3. The line intersects the sphere twice
	/// \param	p1		-	This is used to store the first intersect point. If the line does not
	///						intersect the sphere then ignore this value
	/// \param	p2		-	This is used to store the second intersect point. If the line does not
	///						intersect the sphere, or only one intersect exists then ignore this value
	/// \param	line	-	the line to test for intersection with this sphere
	/// \return 0 if no intersections exists
	///			1 if one intersection point
	///			2 if 2 intersections.
	///
unsigned int XSphere::GetLineIntersects(OUT XVector& p1,OUT XVector& p2,IN const XLine& line) const
	{
		bool IsPoint1InSphere = false;
		bool IsPoint2InSphere = false;

		// check both end points to see if either is in the sphere
		if( PointInSphere( line.Point1 ) )
			IsPoint1InSphere = true;
		if( PointInSphere( line.Point2 ) )
			IsPoint2InSphere = true;

		// if both points in line are in the sphere return 0....
		if( IsPoint1InSphere && IsPoint2InSphere )
			return 0;

		// Create the vector from end point Point1 to our point p.
		XVector v1 = Center - line.Point1;

		// Create a direction vector from end point Point1 to end point Point2
		XVector Dir = (line.Point2 - line.Point1);

		// calculate distance between end points
		float d = Dir.Length();

		// normalise the direction vector
		Dir.Normalise();

		// Using the dot product, we project the v1 onto the vector v2.
		// This essentially gives us the distance from our projected vector from Point1.
		float t = Dir * v1;

		// Here we create a vector that is of length t and in the direction of vVector2
		XVector ClosestPointOnLine = line.Point1 + Dir * t;

		// the line does not intersect the sphere
		if( !PointInSphere(ClosestPointOnLine) )
			return 0;

		// now here is the predicament, in the lines implimentation of this closest point function
		// ensures that the point is on the line by clamping the result to either of the two end points if
		// it goes out of range. This cannot be done here so we need to perform some additional test to see
		// if our possibly imaginary line actually does intersect the sphere.
		if( !IsPoint1InSphere && !IsPoint2InSphere )
		{
			// if neither end points are in sphere then we must make sure that this point lies along
			// the line between them, if not than our line does not intersect the sphere.
			if(t <= 0 || t >= d)
				return 0;
		}

		// now, the next step is to make a bit of a triangle. We now that the line between
		// the closest point on the line and the spheres center is 90 degrees to the line.
		// we can easily find out the length of this line. We also know that the line must intersect
		// the sphere. At this point there is also a straight line to the center of the sphere.
		// This means that we can use pythagorus to determine the length of the third side
		//

		XReal lenFromCpToCenter = (Center-ClosestPointOnLine).Length();
		XReal thirdSideLen = static_cast<XReal>( sqrt( lenFromCpToCenter*lenFromCpToCenter + Radius*Radius ) );

		int numIsects=0;

		// if point 1 is in sphere, ignore it....
		if( !IsPoint1InSphere )
		{
			++numIsects;
			p1 = ClosestPointOnLine - Dir*thirdSideLen;
		}

		// if point 2 is in sphere, ignore it....
		if( !IsPoint2InSphere )
		{
			XVector is2 = ClosestPointOnLine + Dir*thirdSideLen;
			if(numIsects==0)
				p1 = is2;
			else
				p2 = is2;
			++numIsects;
		}

		return numIsects;
	}




