
#ifdef WIN32
#pragma once
#endif


#ifndef TRANSFORM_H_INCLUDED
#define TRANSFORM_H_INCLUDED

// crappy VC6 template warnings
#ifdef WIN32
#pragma warning(disable:4786)
#endif
#include <vector>
#include "./Vector.h"
#include "./Matrix.h"
#include "./Plane.h"
#include "./Quaternion.h"

///-----------------------------------------------------------------------------
/// \brief	The transform class holds scale, translation and rotation information. 
///			It is primarily used as a structure to hold a keyframe for an animation
///			cycle. 
/// 
class XTransform {

	/// stream extraction operator
	friend std::istream& operator>>(std::istream&,XTransform &);

	/// \brief	This function linearly interpolates the transform, however it 
	///			performs a Sphereical interpolation on the rotation components.
	/// \param	out	-	the output transform
	/// \param	t	-	the interpolation value
	/// \param	t1	-	the first transform
	/// \param	t2	-	the second transform
	/// 
	friend void LERP(OUT XTransform&,IN const float,IN const XTransform&,IN const XTransform&);

public:

	const XTransform& operator = (const XTransform& T) {
		m_Rotation = T.m_Rotation;
		m_Translation = T.m_Translation;
		m_Scale = T.m_Scale;
		return *this;
	}


	///-------------------------------------------------------------------------
	/// \brief	ctor
	/// 
	XTransform()
		: m_Rotation(),
		  m_Translation(),
		  m_Scale(1.0f,1.0f,1.0f)
	{
	}

	///-------------------------------------------------------------------------
	/// \brief	A slightly more intersting ctor
	/// \param	r	-	the rotation quaternion for this transform
	/// \param	t	-	the translation for this transform
	/// \param	s	-	the scale for this transform
	/// 
	XTransform(IN const XQuaternion& r,
			   IN const XVector& t,
			   IN const XVector& s )
		: m_Rotation(r),
		  m_Translation(t),
		  m_Scale(s)
	{
	}

	///-------------------------------------------------------------------------
	/// \brief	The copy constructor for the transform types
	/// \param	rhs	-	the copied argument
	/// 
	XTransform(IN const XTransform& rhs)
		: m_Rotation(rhs.m_Rotation),
		  m_Translation(rhs.m_Translation),
		  m_Scale(rhs.m_Scale)
	{
	}

	///-------------------------------------------------------------------------
	/// \brief	This function returns the last calculated bone rotation
	/// \return	the rotation value
	/// 
	inline const XQuaternion& Rotation() const {
		return m_Rotation;
	}

	///-------------------------------------------------------------------------
	/// \brief	This function returns the last calculated bone rotation
	/// \return	the rotation value
	/// 
	inline XQuaternion& Rotation() {
		return m_Rotation;
	}

	///-------------------------------------------------------------------------
	/// \brief	This function returns the last calculated bone translation
	/// \return	the translation value
	/// 
	inline const XVector& Translation() const {
		return m_Translation;
	}

	///-------------------------------------------------------------------------
	/// \brief	This function returns the last calculated bone translation
	/// \return	the translation value
	/// 
	inline XVector& Translation() {
		return m_Translation;
	}


	///-------------------------------------------------------------------------
	/// \brief	This function returns the last calculated bone scale
	/// \return	the scale value
	/// 
	inline const XVector& Scale() const {
		return m_Scale;
	}

	///-------------------------------------------------------------------------
	/// \brief	This function returns the last calculated bone scale
	/// \return	the scale value
	/// 
	inline XVector& Scale() {
		return m_Scale;
	}

	inline void AsMatrix(XMatrix& M) const {
		M.Identity();
		M.Translate(m_Translation);
		XMatrix TEMP;
		m_Rotation.AsMatrix(TEMP.data);
		M*=TEMP;
		M.Scale(m_Scale);
	}


	/// the current value of the rotation
	XQuaternion					m_Rotation;
	
	/// the current value of the translation
	XVector						m_Translation;

	/// the current value of the scale
	XVector						m_Scale;
};


/// a list of lights type
typedef std::vector< XTransform > XTransformList;

/// an iterator for a light list
typedef XTransformList::iterator XItTransform;

#endif
