//=========================================================================
///
///	\file	Vector.cpp
///	\brief	impliments the class XVector, a class definition for a 3d vector
///
//=========================================================================


//=========================================================================
//	Includes
//=========================================================================

#include "Vector.h"

//=========================================================================
//	XVector implimentation
//=========================================================================


//=========================================================================
/// \brief	The constructor for the Vec3 class. This
///			is a special function that gets called
///			whenever the class gets instantaited. This
///			allows you to set default values.
///			This example initialises x,y & z to zero.
XVector::XVector() : x(0), y(0), z(0), w(static_cast<XReal>(1.0f))
{
}


//=========================================================================
/// \brief	The constructor for the Vec3 class that accepts
///			an array of doubles as the default argument.
///
XVector::XVector(const float*  const v)
	: x(static_cast<XReal>(v[0])), 
	  y(static_cast<XReal>(v[1])), 
	  z(static_cast<XReal>(v[2])),
	  w(static_cast<XReal>(1.0f))
{
}


//=========================================================================
/// \brief	The constructor for the Vec3 class that accepts
///			an array of floats as the default argument.
///
XVector::XVector(const double*  const v) 
	: x(static_cast<XReal>(v[0])), 
	  y(static_cast<XReal>(v[1])), 
	  z(static_cast<XReal>(v[2])),
	  w(static_cast<XReal>(1.0f))
{
}


//=========================================================================
/// \brief	The copy constructor for the Vec3 class.
///
XVector::XVector(const XVector &v) : x(v.x), y(v.y), z(v.z), w(v.w)
{
}


//=========================================================================
/// \brief	The constructor for the Vec3 class. This function
///			accepts three floats as it's arguments.
///
XVector::XVector( const XReal x_ ,
			  const XReal y_ ,
			  const XReal z_ ,
			  const XReal w_ )
{
	x = static_cast<XReal>(x_);
	y = static_cast<XReal>(y_);
	z = static_cast<XReal>(z_);
	w = static_cast<XReal>(w_);
}




//=========================================================================
/// \brief	The destructor gets called whenever the object
///			is deleted and allows us to do things like
///			de-allocate memory etc.
///			This example does nothing.....
XVector::~XVector()
{
}



XVector XVector::Orthogonalise(const XVector& v1, const XVector& v2)
{
	XVector v2ProjV1 = ClosestPointOnLine(v1, -v1, v2);
	XVector res = v2 - v2ProjV1;
	res.Normalise();
	return res;
}
XVector XVector::ClosestPointOnLine(const XVector & a,const  XVector & b,const  XVector & p)
{
	// a-b is the line, p the point in question
	XVector c = p-a;
	XVector V = b-a;
	float d = V.Length();
	V = V.Normalise(); // normalize V
	float t = V * c; // V dot c
	// Check to see if t is beyond the extents of the line segment
	if (t < 0.0f)
	{
		XVector res = a;
		return res;
	}
	if (t > d)
	{
		XVector res = b;
		return res;
	}
	// Return the point between a and b
	V = V*t; //set length of V to t.
	XVector res = a + V;
	return res;
}

/// a function to spherically interpolate the two quaternions
//
XVector SLERP(const XVector &q1,const XVector &q2,const float &t)
{
	// for some reason, though why I don't actually know, the second quaternion
	// needs to be inverted if the angle between them is greater than 90 begrees.
	// I think it is to control the direction of the angular rotation so that the
	// bone always slerps through the shortest angle. Just guessing though so don't
	// take my word for it....
	//
	XVector Second;

	// will hold the final position
	XVector Final;

	// if two quarternions are the same
	if( FCOMPARE( q1.x, q2.x ) &&
		FCOMPARE( q1.y, q2.y ) &&
		FCOMPARE( q1.z, q2.z ) )
	{
		// no point in slerping them....
		return q1;
	}

	// perform dot product
	float product = ( q1.x * q2.x ) +
		( q1.y * q2.y ) +
		( q1.z * q2.z ) +
		( q1.w * q2.w );

	// if the angle is greater than 90degress, negate quaternion and product
	if ( product < 0.001f )
	{
		Second  = XVector(-q2.x,-q2.y,-q2.z);
		product = -product;
	}
	else
	{
		Second = q2;
	}

	// calculate the linear t values for the interpolation
	float interpolating_t    = t;
	float interpolating_invt = 1.0f-t;

	if(product > 1) 
		product=0.999999f;

	if(product < -1) 
		product=-0.999999f;

	// spherical interpolation takes place around a sphere. Therefore we use the
	// product (remeber dot product to finds angle between 2 vectors) to get the
	// angle we have to rotate to go from q1 to q2.
	//
	float theta    = static_cast<float>( acos(product) );

	// precalculate the sin value
	//
	float sinTheta = static_cast<float>( sin(theta) );

	if(sinTheta != 0.0f)
	{
		interpolating_invt = static_cast<float>( sin(interpolating_invt * theta) / sinTheta   );
		interpolating_t    = static_cast<float>( sin(interpolating_t    * theta) / sinTheta   );
	}

	// perform the interpolation
	Final.x = (interpolating_invt * q1.x) + (interpolating_t * Second.x);
	Final.y = (interpolating_invt * q1.y) + (interpolating_t * Second.y);
	Final.z = (interpolating_invt * q1.z) + (interpolating_t * Second.z);

	Final.Normalise();

	// return the result of interpolation
	return Final;
}








