//=========================================================================
///
///	\file	Vector.h
/// \author	Rob
///	\brief	Defines the class XVector, a class definition for a 3d vector
///
//=========================================================================



//=========================================================================
//	Sentry
//=========================================================================

//	This directive is used to stop the header file from being included
//	Twice and is standard for all headers.
#ifdef WIN32
#pragma once
#endif


#ifndef VECTOR_H_INCLUDED
#define VECTOR_H_INCLUDED


//=========================================================================
//	Includes
//=========================================================================

#include <math.h>
#include <iostream>
#include <vector>
#include <assert.h>
#include "MathDefs.h"


//=========================================================================
//	Class Definitions
//=========================================================================

/// \brief	a small class to hold a 3d vector
class XVector
{
	friend XVector SLERP(const XVector&,const XVector&,const float&);
public:


	///	\brief	this union is used to allow two representations of the data
	///			contained within the class.
	union
	{
		/// \brief	a nameless struct to bolt together 4 floats
		///
		struct
		{
			///	The class member data as individual components
			///
			XReal	x,
					y,
					z,
					w;
		};
		///	The class member data as an array
		///
		XReal data[4];
	};


	/// \brief	The constructor for the Vec3 class. This
	///			is a special function that gets called
	///			whenever the class gets instantaited. This
	///			allows you to set default values.
	XVector();

	/// \brief	The constructor for the Vec3 class that accepts
	///			an array of floats as the default argument.
	/// \param	v	-	an array of three floats
	///
	XVector(IN const float*  const v);

	/// \brief	The constructor for the Vec3 class that accepts
	///			an array of floats/doubles as the default argument.
	/// \param	d	-	an array of three doubles
	///
	XVector(IN const double* const d);

	/// \brief	The copy constructor for the Vec3 class.
	/// \param	v	-	the XVector instance to copy.
	///
	XVector(IN const XVector& v);

	/// \brief	The constructor for the Vec3 class. This function
	///			accepts three/four floats as it's arguments.
	/// \param	x	-	the x component value
	/// \param	y	-	the y component value
	/// \param	z	-	the z component value
	/// \param	w	-	the (optional) w component value
	///
	XVector(IN const XReal,
		  IN const XReal,
		  IN const XReal,
		  IN const XReal = 1.0f);

	/// \brief	The destructor gets called whenever the object
	///			is deleted and allows us to do things like
	///			de-allocate memory etc.
	~XVector();


	//	STREAM INSERTION/EXTRACTION

	///	\brief	operator  std::istream >> XVector
	///	\param	ifs	-	the input file stream, ie std::cin
	///	\param	v	-	the vector instance into which the data is being placed.
	/// \return	the stream reference
	/// \note	The stream extraction operator has to be declared as a friend of the
	///			class. This is because the operator itself cannot be a member function
	///			of XVector because the type, std::istream, appears on the left of the
	///			operator. For it to be a member function, the CMatrix would have to be
	///			on the left of the >> operator. This forces us to make the operator act
	///			globally and should therefore always be declared as a friend.
	///
	friend std::istream &operator >> (IN std::istream &ifs,IN XVector &v) {
		return ifs >> v.x >> v.y >> v.z;
	}

	///	\brief	operator  std::ostream << XVector
	///	\param	ofs	-	the output file stream, ie std::cout, std::fstream etc
	///	\param	v	-	the vector to output
	/// \return	a reference to the stream
	/// \note	The stream insertion operator has to be declared as a friend of the
	///			class. This is because the operator itself cannot be a member function
	///			of XVector because the type, std::ostream, appears on the left of the
	///			operator. For it to be a member function, the CMatrix would have to be
	///			on the left of the << operator. This forces us to make the operator act
	///			globally and should therefore always be declared as a friend.
	///
	///
	friend std::ostream &operator << (IN std::ostream &ofs,IN const XVector &v) {
		ofs << v.x << " " << v.y << " " << v.z;
		return ofs;
	}



	//	UNARY NEGATION

	///	\brief	operator -	XVector		(unary)
	/// \return	it's opposite so to speak....
	///
	inline OUT XVector operator - () const
	{
		return XVector( -x, -y, -z );
	}



	//	CONDITIONAL OPERATORS

	///	\brief	operator == XVector
	///	\param	v	-	compare v against *this
	/// \return	true if (more or less within accuracy of g_vecEpsilon)
	///			the same as this, false otherwise
	///
	inline const bool operator == (IN const XVector &v) const
	{
		return  (x + g_vecEpsilon) > v.x  &&
				(x - g_vecEpsilon) < v.x  &&
			    (y + g_vecEpsilon) > v.y  &&
				(y - g_vecEpsilon) < v.y  &&
			    (z + g_vecEpsilon) > v.z  &&
				(z - g_vecEpsilon) < v.z  ;
	}

	///	\brief	operator != XVector
	///	\param	v	-	compare v against *this
	/// \return	false if (more or less within accuracy of g_vecEpsilon)
	///			the same as this
	///
	inline const bool operator != (IN const XVector &v) const
	{
		return !(v == *this);
	}



	//	ASSIGNMENT OPERATORS

	///	\brief	operator =	XVector
	///	\param	rhs	-	the vector on the right hand side of the operator
	/// \return	a constant reference to this
	///
	inline const XVector& operator = (IN const XVector &rhs)
	{
		x = rhs.x;
		y = rhs.y;
		z = rhs.z;
		return *this;
	}

	///	\brief	operator +=	XVector
	///	\param	rhs	-	the vector on the right hand side of the operator
	/// \return	a constant reference to this
	///
	inline const XVector& operator += (IN const XVector &rhs)
	{
		x += rhs.x;
		y += rhs.y;
		z += rhs.z;
		return *this;
	}

	///	\brief	operator -=	XVector
	///	\param	rhs	-	the vector on the right hand side of the operator
	/// \return	a constant reference to this
	///
	inline const XVector& operator -= (IN const XVector &rhs)
	{
		x -= rhs.x;
		y -= rhs.y;
		z -= rhs.z;
		return *this;
	}

	///	\brief	operator *=	double
	///	\param	d	-	the double to multiple this by
	/// \return	a constant reference to this
	///
	inline const XVector& operator *= (IN const XReal &d)
	{
		x *= static_cast<XReal>(d);
		y *= static_cast<XReal>(d);
		z *= static_cast<XReal>(d);
		return *this;
	}


	///	\brief	operator /=	float
	///	\param	f	-	the float to divide the vector by
	/// \return	a constant reference to this
	///
	inline const XVector& operator /= (IN const XReal &f)
	{
		x /= static_cast<XReal>(f);
		y /= static_cast<XReal>(f);
		z /= static_cast<XReal>(f);
		return *this;
	}




	//	OTHER BINARY OPERATORS

	///	\brief	operator +	XVector
	///	\param	rhs	-	the vector on the right hand side of the operator
	/// \return	( *this + rhs )
	///
	inline XVector operator + (IN const XVector &rhs) const
	{
		return XVector( x + rhs.x ,
					  y + rhs.y ,
					  z + rhs.z );
	}

	///	\brief	operator -	XVector
	///	\param	rhs	-	the vector on the right hand side of the operator
	/// \return	( *this - rhs )
	///
	inline XVector operator - (IN const XVector &rhs) const
	{
		return XVector( x - rhs.x ,
					  y - rhs.y ,
					  z - rhs.z );
	}

	///	\brief	operator *	XVector		(DOT_PRODUCT)
	///	\param	rhs	-	the vector on the right hand side of the operator
	/// \return	( *this DOT rhs )
	///
	inline const XReal operator * (IN const XVector &rhs) const
	{
		return  x * rhs.x +
				y * rhs.y +
				z * rhs.z;
	}

	///	\brief	operator ^	XVector		(CROSS_PRODUCT)
	///	\param	rhs	-	the vector on the right hand side of the operator
	/// \return	( *this CROSS rhs )
	///
	inline XVector operator ^ (IN  const XVector &v ) const
	{
		return XVector( ( y * v.z  -  z * v.y ),
					  ( z * v.x  -  x * v.z ),
					  ( x * v.y  -  y * v.x ) );
	}

	///	\brief	operator *	float
	///	\param	f	-	the float to scale the vector by
	/// \return	( (*this) * f )
	///
	inline XVector operator * (IN const XReal &f) const
	{
		return XVector( x * static_cast<XReal>(f) ,
					  y * static_cast<XReal>(f) ,
					  z * static_cast<XReal>(f) );
	}
	friend inline XVector operator * (IN const XReal &f, IN const XVector& v)
	{
		return XVector( v.x * static_cast<XReal>(f) ,
					  v.y * static_cast<XReal>(f) ,
					  v.z * static_cast<XReal>(f) );
	}


	///	\brief	operator /	float
	///	\param	f	-	the float to divide by
	/// \return	( (*this) / f )
	///
	inline XVector operator / (IN const float &f)
	{
		return XVector( x / static_cast<XReal>(f) ,
					  y / static_cast<XReal>(f) ,
					  z / static_cast<XReal>(f) );
	}

	///	\brief	operator /	double
	///	\param	d	-	the double to divide by
	/// \return	( (*this) / d )
	///
	inline XVector operator / (IN const double &d)
	{
		return XVector( x / static_cast<XReal>(d) ,
					  y / static_cast<XReal>(d) ,
					  z / static_cast<XReal>(d) );
	}



	/// \brief	a member function to return the length of the vector
	/// \return	the length of the vector
	///
	inline OUT XReal Length() const
	{
		return (XReal) sqrt( (x * x) + (y * y) + (z * z) );
	}



	/// \brief	This function normalises the vector
	/// \return \e nothing
	///
	inline XVector& Normalise()
	{
		// multiplication is quicker than division...
		XReal invL = 1.0f / Length();
		x *= invL;
		y *= invL;
		z *= invL;
		w  = 1.0f;
		return *this;
	}



	/// \brief	This function sets the three values
	/// \param	x_	The x component of the new vector
	/// \param	y_	The y component of the new vector
	/// \param	z_	The z component of the new vector
	/// \return \e nothing
	///
	inline XVector& Set(IN const XReal x_,
					    IN const XReal y_,
					    IN const XReal z_,
					    IN const XReal w_ = static_cast<XReal>(1.0f) )
	{
		x = x_;
		y = y_;
		z = z_;
		w = w_;
		return *this;
	}


	XVector Orthogonalise(const XVector& v1,const XVector& v2); 
	XVector ClosestPointOnLine(const XVector & a,const XVector & b,const XVector & p);

};


///-----------------------------------------------------------------------------
/// \brief	This function linearly interpolates two vectors, v1 & v2, with
///			respect to t
/// \param	v1	-	the first vector to use
/// \param	v2	-	the second vector to use
///	\param	t	-	the interpolation value
///
inline OUT XVector LERP(IN const XVector& v1,
						IN const XVector& v2,
						IN const XReal &t)
{
	XReal invt = 1.0f-t;
	return XVector( ((v1.x*invt)+(v2.x*t)),
					((v1.y*invt)+(v2.y*t)),
					((v1.z*invt)+(v2.z*t)) );
}

typedef std::vector<XVector> XVecList;
typedef XVecList::iterator XItVec;
typedef XVecList::const_iterator XCItVec;

#endif
